# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Helper functions for the local executor
"""

import io
import json
import os
import sys
import traceback
from contextlib import redirect_stderr, redirect_stdout
from pathlib import Path
from typing import Any, Callable, Dict, List, Tuple

import requests

from covalent._shared_files.utils import get_qelectron_db_path
from covalent._workflow.depsbash import DepsBash
from covalent._workflow.depscall import RESERVED_RETVAL_KEY__FILES, DepsCall
from covalent._workflow.depspip import DepsPip
from covalent._workflow.transport import TransportableObject
from covalent.executor.utils import set_context
from covalent.executor.utils.serialize import deserialize_node_asset, serialize_node_asset


def wrapper_fn(
    function: TransportableObject,
    call_before: List[Tuple[TransportableObject, TransportableObject, TransportableObject]],
    call_after: List[Tuple[TransportableObject, TransportableObject, TransportableObject]],
    *args,
    **kwargs,
):
    """Wrapper for serialized callable.

    Execute preparatory shell commands before deserializing and
    running the callable. This is the actual function to be sent to
    the various executors.

    """

    cb_retvals = {}
    for tup in call_before:
        serialized_fn, serialized_args, serialized_kwargs, retval_key = tup
        cb_fn = serialized_fn.get_deserialized()
        cb_args = serialized_args.get_deserialized()
        cb_kwargs = serialized_kwargs.get_deserialized()
        retval = cb_fn(*cb_args, **cb_kwargs)

        # we always store cb_kwargs dict values as arrays to factor in non-unique values
        if retval_key and retval_key in cb_retvals:
            cb_retvals[retval_key].append(retval)
        elif retval_key:
            cb_retvals[retval_key] = [retval]

    # if cb_retvals key only contains one item this means it is a unique (non-repeated) retval key
    # so we only return the first element however if it is a 'files' kwarg we always return as a list
    cb_retvals = {
        key: value[0] if len(value) == 1 and key != RESERVED_RETVAL_KEY__FILES else value
        for key, value in cb_retvals.items()
    }

    fn = function.get_deserialized()

    new_args = [arg.get_deserialized() for arg in args]

    new_kwargs = {k: v.get_deserialized() for k, v in kwargs.items()}

    # Inject return values into kwargs
    for key, val in cb_retvals.items():
        new_kwargs[key] = val

    output = fn(*new_args, **new_kwargs)

    for tup in call_after:
        serialized_fn, serialized_args, serialized_kwargs, retval_key = tup
        ca_fn = serialized_fn.get_deserialized()
        ca_args = serialized_args.get_deserialized()
        ca_kwargs = serialized_kwargs.get_deserialized()
        ca_fn(*ca_args, **ca_kwargs)

    return TransportableObject(output)


def io_wrapper(
    fn: Callable,
    args: List,
    kwargs: Dict,
    workdir: str = ".",
) -> Tuple[Any, str, str, str]:
    """Wrapper function to execute the given function in a separate
    process and capture stdout and stderr"""
    with redirect_stdout(io.StringIO()) as stdout, redirect_stderr(io.StringIO()) as stderr:
        try:
            Path(workdir).mkdir(parents=True, exist_ok=True)
            current_dir = os.getcwd()
            os.chdir(workdir)
            output = fn(*args, **kwargs)
            tb = ""
        except Exception as ex:
            output = None
            tb = "".join(traceback.TracebackException.from_exception(ex).format())
        finally:
            os.chdir(current_dir)
    return output, stdout.getvalue(), stderr.getvalue(), tb


# Copied from runner.py
def _gather_deps(deps, call_before_objs_json, call_after_objs_json) -> Tuple[List, List]:
    """Assemble deps for a node into the final call_before and call_after"""

    call_before = []
    call_after = []

    # Rehydrate deps from JSON
    if "bash" in deps:
        dep = DepsBash()
        dep.from_dict(deps["bash"])
        call_before.append(dep.apply())

    if "pip" in deps:
        dep = DepsPip()
        dep.from_dict(deps["pip"])
        call_before.append(dep.apply())

    for dep_json in call_before_objs_json:
        dep = DepsCall()
        dep.from_dict(dep_json)
        call_before.append(dep.apply())

    for dep_json in call_after_objs_json:
        dep = DepsCall()
        dep.from_dict(dep_json)
        call_after.append(dep.apply())

    return call_before, call_after


# Basic wrapper for executing a topologically sorted sequence of
# tasks. For the `task_specs` and `resources` schema see the comments
# for `AsyncBaseExecutor.send()`.


def run_task_group(
    task_specs: List[Dict],
    output_uris: List[Tuple[str, str, str]],
    results_dir: str,
    task_group_metadata: dict,
    server_url: str,
):
    """
    Run a task group.

    This is appropriate for executors which can access the Covalent
    server url directly. Exampl: LocalExecutor.

    """

    prefix = "file://"
    prefix_len = len(prefix)

    outputs = {}
    results = []
    dispatch_id = task_group_metadata["dispatch_id"]
    task_ids = task_group_metadata["node_ids"]
    gid = task_group_metadata["task_group_id"]

    os.environ["COVALENT_DISPATCH_ID"] = dispatch_id
    os.environ["COVALENT_DISPATCHER_URL"] = server_url

    for i, task in enumerate(task_specs):
        result_uri, stdout_uri, stderr_uri, qelectron_db_uri = output_uris[i]

        # Setting these to empty bytes in case the task fails
        qelectron_db_bytes = bytes()

        with open(stdout_uri, "w") as stdout, open(stderr_uri, "w") as stderr:
            with redirect_stdout(stdout), redirect_stderr(stderr):
                try:
                    task_id = task["function_id"]
                    args_ids = task["args_ids"]
                    kwargs_ids = task["kwargs_ids"]

                    function_uri = f"{server_url}/api/v2/dispatches/{dispatch_id}/electrons/{task_id}/assets/function"

                    # Download function
                    resp = requests.get(function_uri, stream=True)
                    resp.raise_for_status()
                    serialized_fn = deserialize_node_asset(resp.content, "function")

                    ser_args = []
                    ser_kwargs = {}

                    # Download args and kwargs
                    for node_id in args_ids:
                        url = f"{server_url}/api/v2/dispatches/{dispatch_id}/electrons/{node_id}/assets/output"
                        resp = requests.get(url, stream=True)
                        resp.raise_for_status()
                        ser_args.append(deserialize_node_asset(resp.content, "output"))

                    for k, node_id in kwargs_ids.items():
                        url = f"{server_url}/api/v2/dispatches/{dispatch_id}/electrons/{node_id}/assets/output"
                        resp = requests.get(url, stream=True)
                        resp.raise_for_status()
                        ser_kwargs[k] = deserialize_node_asset(resp.content, "output")

                    # Download deps, call_before, and call_after
                    hooks_url = f"{server_url}/api/v2/dispatches/{dispatch_id}/electrons/{task_id}/assets/hooks"
                    resp = requests.get(hooks_url, stream=True)
                    resp.raise_for_status()
                    hooks_json = deserialize_node_asset(resp.content, "hooks")
                    deps_json = hooks_json.get("deps", {})
                    call_before_json = hooks_json.get("call_before", [])
                    call_after_json = hooks_json.get("call_after", [])

                    # Assemble and run the task
                    call_before, call_after = _gather_deps(
                        deps_json, call_before_json, call_after_json
                    )
                    exception_occurred = False

                    with set_context(dispatch_id, task_id):
                        transportable_output = wrapper_fn(
                            serialized_fn, call_before, call_after, *ser_args, **ser_kwargs
                        )

                    ser_output = serialize_node_asset(transportable_output, "output")
                    with open(result_uri, "wb") as f:
                        f.write(ser_output)

                    qelectron_db_path = get_qelectron_db_path(dispatch_id, task_id)
                    if qelectron_db_path is not None:
                        with open(qelectron_db_path / "data.mdb", "rb") as f:
                            qelectron_db_bytes = f.read()

                    outputs[task_id] = result_uri

                    result_summary = {
                        "node_id": task_id,
                        "output_uri": result_uri,
                        "stdout_uri": stdout_uri,
                        "stderr_uri": stderr_uri,
                        "qelectron_db_uri": qelectron_db_uri,
                        "exception_occurred": exception_occurred,
                    }

                except Exception as ex:
                    exception_occurred = True
                    tb = "".join(traceback.TracebackException.from_exception(ex).format())
                    print(tb, file=sys.stderr)
                    result_uri = None
                    result_summary = {
                        "node_id": task_id,
                        "output_uri": result_uri,
                        "stdout_uri": stdout_uri,
                        "stderr_uri": stderr_uri,
                        "qelectron_db_uri": qelectron_db_uri,
                        "exception_occurred": exception_occurred,
                    }

                    break

                finally:
                    # POST task artifacts
                    if result_uri:
                        upload_url = f"{server_url}/api/v2/dispatches/{dispatch_id}/electrons/{task_id}/assets/output"
                        with open(result_uri, "rb") as f:
                            requests.put(upload_url, data=f)

                    sys.stdout.flush()
                    if stdout_uri:
                        upload_url = f"{server_url}/api/v2/dispatches/{dispatch_id}/electrons/{task_id}/assets/stdout"
                        with open(stdout_uri, "rb") as f:
                            headers = {"Content-Length": os.path.getsize(stdout_uri)}
                            requests.put(upload_url, data=f)

                    sys.stderr.flush()
                    if stderr_uri:
                        upload_url = f"{server_url}/api/v2/dispatches/{dispatch_id}/electrons/{task_id}/assets/stderr"
                        with open(stderr_uri, "rb") as f:
                            headers = {"Content-Length": os.path.getsize(stderr_uri)}
                            requests.put(upload_url, data=f)

                    if qelectron_db_bytes:
                        upload_url = f"{server_url}/api/v2/dispatches/{dispatch_id}/electrons/{task_id}/assets/qelectron_db"
                        requests.put(upload_url, data=qelectron_db_bytes)

                    result_path = os.path.join(results_dir, f"result-{dispatch_id}:{task_id}.json")

                    with open(result_path, "w") as f:
                        json.dump(result_summary, f)

                    results.append(result_summary)

                    # Notify Covalent that the task has terminated
                    terminal_status = "FAILED" if exception_occurred else "COMPLETED"
                    url = f"{server_url}/api/v2/dispatches/{dispatch_id}/electrons/{task_id}/job"
                    data = {"status": terminal_status}
                    requests.put(url, json=data)

    # Deal with any tasks that did not run
    n = len(results)
    if n < len(task_ids):
        for i in range(n, len(task_ids)):
            result_summary = {
                "node_id": task_ids[i],
                "output_uri": "",
                "stdout_uri": "",
                "stderr_uri": "",
                "qelectron_db_uri": "",
                "exception_occurred": True,
            }

            results.append(result_summary)

            result_path = os.path.join(results_dir, f"result-{dispatch_id}:{task_id}.json")

            with open(result_path, "w") as f:
                json.dump(result_summary, f)

            url = f"{server_url}/api/v2/dispatches/{dispatch_id}/electrons/{task_id}/job"
            requests.put(url)


def run_task_group_alt(
    task_specs: List[Dict],
    resources: dict,
    output_uris: List[Tuple[str, str, str]],
    results_dir: str,
    task_group_metadata: dict,
    server_url: str,
):
    """
    Alternate form of run_task_group.

    This is appropriate for backends that cannot reach the Covalent
    server. Covalent will push input assets to the executor's
    persistent storage before invoking `Executor.send()` and pull output
    artifacts after `Executor.receive()`.

    Example: DaskExecutor.

    """

    prefix = "file://"
    prefix_len = len(prefix)

    outputs = {}
    results = []
    dispatch_id = task_group_metadata["dispatch_id"]
    task_ids = task_group_metadata["node_ids"]
    gid = task_group_metadata["task_group_id"]

    os.environ["COVALENT_DISPATCH_ID"] = dispatch_id
    os.environ["COVALENT_DISPATCHER_URL"] = server_url

    for i, task in enumerate(task_specs):
        result_uri, stdout_uri, stderr_uri, qelectron_db_uri = output_uris[i]

        # Setting these to empty bytes in case the task fails
        qelectron_db_bytes = bytes()

        with open(stdout_uri, "w") as stdout, open(stderr_uri, "w") as stderr:
            with redirect_stdout(stdout), redirect_stderr(stderr):
                try:
                    task_id = task["function_id"]
                    args_ids = task["args_ids"]
                    kwargs_ids = task["kwargs_ids"]

                    # Load function
                    function_uri = resources["functions"][task_id]
                    if function_uri.startswith(prefix):
                        function_uri = function_uri[prefix_len:]

                    with open(function_uri, "rb") as f:
                        serialized_fn = deserialize_node_asset(f.read(), "function")

                    # Load args and kwargs
                    ser_args = []
                    ser_kwargs = {}

                    args_uris = [resources["inputs"][index] for index in args_ids]
                    for uri in args_uris:
                        if uri.startswith(prefix):
                            uri = uri[prefix_len:]
                        with open(uri, "rb") as f:
                            ser_args.append(deserialize_node_asset(f.read(), "output"))

                    kwargs_uris = {k: resources["inputs"][v] for k, v in kwargs_ids.items()}
                    for key, uri in kwargs_uris.items():
                        if uri.startswith(prefix):
                            uri = uri[prefix_len:]
                        with open(uri, "rb") as f:
                            ser_kwargs[key] = deserialize_node_asset(f.read(), "output")

                    # Load deps, call_before, and call_after
                    hooks_uri = resources["hooks"][task_id]
                    if hooks_uri.startswith(prefix):
                        hooks_uri = hooks_uri[prefix_len:]
                    with open(hooks_uri, "rb") as f:
                        hooks_json = deserialize_node_asset(f.read(), "hooks")
                    deps_json = hooks_json.get("deps", {})
                    call_before_json = hooks_json.get("call_before", [])
                    call_after_json = hooks_json.get("call_after", [])

                    # Assemble and invoke the task
                    call_before, call_after = _gather_deps(
                        deps_json, call_before_json, call_after_json
                    )

                    exception_occurred = False

                    # Run the task function
                    with set_context(dispatch_id, task_id):
                        transportable_output = wrapper_fn(
                            serialized_fn, call_before, call_after, *ser_args, **ser_kwargs
                        )

                    ser_output = serialize_node_asset(transportable_output, "output")

                    # Save output
                    with open(result_uri, "wb") as f:
                        f.write(ser_output)

                    # Save QElectron DB
                    qelectron_db_path = get_qelectron_db_path(dispatch_id, task_id)
                    if qelectron_db_path is not None:
                        with open(qelectron_db_path / "data.mdb", "rb") as f:
                            qelectron_db_bytes = f.read()

                    with open(qelectron_db_uri, "wb") as f:
                        f.write(qelectron_db_bytes)

                    resources["inputs"][task_id] = result_uri

                    output_size = len(ser_output)
                    qelectron_db_size = len(qelectron_db_bytes)
                    stdout.flush()
                    stderr.flush()
                    stdout_size = os.path.getsize(stdout_uri)
                    stderr_size = os.path.getsize(stderr_uri)
                    result_summary = {
                        "node_id": task_id,
                        "output": {
                            "uri": result_uri,
                            "size": output_size,
                        },
                        "stdout": {
                            "uri": stdout_uri,
                            "size": stdout_size,
                        },
                        "stderr": {
                            "uri": stderr_uri,
                            "size": stderr_size,
                        },
                        "qelectron_db": {
                            "uri": qelectron_db_uri,
                            "size": qelectron_db_size,
                        },
                        "exception_occurred": exception_occurred,
                    }

                except Exception as ex:
                    exception_occurred = True
                    tb = "".join(traceback.TracebackException.from_exception(ex).format())
                    print(tb, file=sys.stderr)
                    stdout.flush()
                    stderr.flush()
                    stdout_size = os.path.getsize(stdout_uri)
                    stderr_size = os.path.getsize(stderr_uri)
                    result_summary = {
                        "node_id": task_id,
                        "output": {
                            "uri": "",
                            "size": 0,
                        },
                        "stdout": {
                            "uri": stdout_uri,
                            "size": stdout_size,
                        },
                        "stderr": {
                            "uri": stderr_uri,
                            "size": stderr_size,
                        },
                        "qelectron_db": {
                            "uri": "",
                            "size": 0,
                        },
                        "exception_occurred": exception_occurred,
                    }

                    break

                finally:
                    results.append(result_summary)
                    result_path = os.path.join(results_dir, f"result-{dispatch_id}:{task_id}.json")

                    # Write the summary file containing the URIs for
                    # the serialized result, stdout, stderr, and qelectron_db
                    with open(result_path, "w") as f:
                        json.dump(result_summary, f)

    # Deal with any tasks that did not run
    n = len(results)
    if n < len(task_ids):
        for i in range(n, len(task_ids)):
            result_summary = {
                "node_id": task_id,
                "output": {
                    "uri": "",
                    "size": 0,
                },
                "stdout": {
                    "uri": "",
                    "size": 0,
                },
                "stderr": {
                    "uri": "",
                    "size": 0,
                },
                "qelectron_db": {
                    "uri": "",
                    "size": 0,
                },
                "exception_occurred": True,
            }

            results.append(result_summary)

            result_path = os.path.join(results_dir, f"result-{dispatch_id}:{task_id}.json")

            with open(result_path, "w") as f:
                json.dump(result_summary, f)



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from os import path
from pathlib import Path
from subprocess import PIPE, Popen

import click

from ..._db.datastore import DataStore

MIGRATION_WARNING_MSG = "There was an issue running migrations.\nPlease read https://covalent.readthedocs.io/en/latest/how_to/db/migration_error.html for more information."


@click.group(invoke_without_command=True)
@click.pass_context
def db(ctx: click.Context):
    """
    Manage server's dispatch database
    """
    if ctx.invoked_subcommand is None:
        click.echo(ctx.get_help())


@click.command()
@click.pass_context
def migrate(ctx: click.Context) -> None:
    """
    Run database migrations
    """
    try:
        db = DataStore.factory()
        db.run_migrations()
        click.secho("Migrations are up to date.", fg="green")
    except Exception as migration_error:
        click.echo(str(migration_error))
        click.secho(MIGRATION_WARNING_MSG, fg="red")
        return ctx.exit(1)


@click.command(
    context_settings=dict(
        ignore_unknown_options=True,
    )
)
@click.pass_context
@click.argument("alembic_args", nargs=-1, type=click.UNPROCESSED)
def alembic(ctx: click.Context, alembic_args) -> None:
    """
    Alembic CLI
    """
    try:
        alembic_args = list(alembic_args)
        migrations_folder_path = Path(path.join(__file__, "./../../../../covalent_migrations/"))
        project_root_path = migrations_folder_path / Path("..")
        settings_file_path = migrations_folder_path / Path("alembic.ini")
        alembic_command = ["alembic", "-c", str(settings_file_path.resolve())] + alembic_args
        p = Popen(alembic_command, stdout=PIPE, stderr=PIPE, cwd=str(project_root_path.resolve()))
        output, error = p.communicate()
        if error:
            click.echo(error.decode("utf-8").strip())
        else:
            click.echo(output.decode("utf-8").strip())
    except Exception as migration_error:
        if migration_error:
            click.echo(f"{type(migration_error)}:{str(migration_error)}")
        click.secho(
            "There was an error forwarding arguments to alembic CLI please ensure that alembic is installed.",
            fg="red",
        )
        return ctx.exit(1)


db.add_command(alembic)
db.add_command(migrate)



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Defines the core functionality of the result service
"""

import asyncio
import tempfile
import traceback
from typing import Dict

from pydantic import ValidationError

from covalent._dispatcher_plugins.local import LocalDispatcher
from covalent._results_manager import Result
from covalent._shared_files import logger
from covalent._shared_files.schemas.result import ResultSchema
from covalent._shared_files.util_classes import RESULT_STATUS
from covalent._workflow.lattice import Lattice

from .._dal.result import Result as SRVResult
from .._dal.result import get_result_object as get_result_object_from_db
from .._db.write_result_to_db import resolve_electron_id
from . import dispatcher
from .data_modules import dispatch, electron  # nopycln: import
from .data_modules import importer as manifest_importer
from .data_modules.utils import run_in_executor

app_log = logger.app_log
log_stack_info = logger.log_stack_info


def generate_node_result(
    node_id: int,
    node_name: str = None,
    start_time=None,
    end_time=None,
    status=None,
    output=None,
    error=None,
    stdout=None,
    stderr=None,
    qelectron_data_exists=None,
):
    """
    Helper routine to prepare the node result

    Arg(s)
        node_id: ID of the node in the trasport graph
        node_name: Name of the node
        start_time: Start time of the node
        end_time: Time at which the node finished executing
        status: Status of the node's execution
        output: Output of the node
        error: Error from the node
        stdout: STDOUT of a node
        stderr: STDERR generated during node execution
        qelectron_data_exists: Whether the qelectron data exists

    Return(s)
        Dictionary of the inputs
    """

    return {
        "node_id": node_id,
        "node_name": node_name,
        "start_time": start_time,
        "end_time": end_time,
        "status": status,
        "output": output,
        "error": error,
        "stdout": stdout,
        "stderr": stderr,
        "qelectron_data_exists": qelectron_data_exists,  # TODO: This field is now defunct, see PR #1850
    }


# Domain: result
async def update_node_result(dispatch_id, node_result):
    app_log.debug("Updating node result (run_planned_workflow).")
    valid_update = True
    try:
        node_id = node_result["node_id"]
        node_status = node_result["status"]
        node_info = await electron.get(dispatch_id, node_id, ["type", "sub_dispatch_id"])
        node_type = node_info["type"]
        sub_dispatch_id = node_info["sub_dispatch_id"]

        # Handle returns from _build_sublattice_graph -- change
        # COMPLETED -> DISPATCHING
        node_result = _filter_sublattice_status(
            dispatch_id, node_id, node_status, node_type, sub_dispatch_id, node_result
        )

        valid_update = await electron.update(dispatch_id, node_result)
        if not valid_update:
            app_log.warning(
                f"Invalid status update {node_status} for node {dispatch_id}:{node_id}"
            )
            return

        if node_result["status"] == RESULT_STATUS.DISPATCHING:
            app_log.debug("Received sublattice dispatch")
            try:
                sub_dispatch_id = await _make_sublattice_dispatch(dispatch_id, node_result)
            except Exception as ex:
                tb = "".join(traceback.TracebackException.from_exception(ex).format())
                node_result["status"] = RESULT_STATUS.FAILED
                node_result["error"] = tb
                await electron.update(dispatch_id, node_result)

    except KeyError as ex:
        valid_update = False
        app_log.exception(f"Error persisting node update: {ex}")

    except Exception as ex:
        app_log.exception(f"Error persisting node update: {ex}")
        sub_dispatch_id = None
        node_result["status"] = Result.FAILED

    finally:
        if not valid_update:
            return

        node_id = node_result["node_id"]
        node_status = node_result["status"]
        detail = {"sub_dispatch_id": sub_dispatch_id} if sub_dispatch_id else {}

        if node_status and valid_update:
            dispatch_id = dispatch_id
            await dispatcher.notify_node_status(dispatch_id, node_id, node_status, detail)


# Domain: result
def _redirect_lattice(
    json_lattice: str,
    parent_dispatch_id: str,
    parent_electron_id: int,
    loop: asyncio.AbstractEventLoop,
) -> str:
    """Redirect a JSON lattice through the new DAL.

    Args:
        json_lattice: A JSON-serialized lattice.
        parent_dispatch_id:  The id of a sublattice's parent dispatch.

    This will only be triggered from either the monolithic /submit
    endpoint or a monolithic sublattice dispatch.

    Returns:
        The dispatch manifest

    """
    lattice = Lattice.deserialize_from_json(json_lattice)
    with tempfile.TemporaryDirectory() as staging_dir:
        manifest = LocalDispatcher.prepare_manifest(lattice, staging_dir)

        # Trigger an internal asset pull from /tmp to object store
        coro = manifest_importer.import_manifest(
            manifest,
            parent_dispatch_id,
            parent_electron_id,
        )
        filtered_manifest = manifest_importer._import_manifest(
            manifest,
            parent_dispatch_id,
            parent_electron_id,
        )

        manifest_importer._pull_assets(filtered_manifest)

    return filtered_manifest.metadata.dispatch_id


async def make_dispatch(
    json_lattice: str, parent_dispatch_id: str = None, parent_electron_id: int = None
) -> str:
    return await run_in_executor(
        _redirect_lattice,
        json_lattice,
        parent_dispatch_id,
        parent_electron_id,
        asyncio.get_running_loop(),
    )


def get_result_object(dispatch_id: str, bare: bool = True) -> SRVResult:
    app_log.debug(f"Getting result object from db, bare={bare}")
    return get_result_object_from_db(dispatch_id, bare)


def finalize_dispatch(dispatch_id: str):
    app_log.debug(f"Finalizing dispatch {dispatch_id}")


async def persist_result(dispatch_id: str):
    await _update_parent_electron(dispatch_id)


async def _update_parent_electron(dispatch_id: str):
    dispatch_attrs = await dispatch.get(dispatch_id, ["electron_id", "status", "end_time"])
    parent_eid = dispatch_attrs["electron_id"]

    if parent_eid:
        dispatch_id, node_id = resolve_electron_id(parent_eid)
        status = dispatch_attrs["status"]
        node_result = generate_node_result(
            node_id=node_id,
            end_time=dispatch_attrs["end_time"],
            status=status,
        )
        parent_result_obj = get_result_object(dispatch_id)
        app_log.debug(f"Updating sublattice parent node {dispatch_id}:{node_id}")
        await update_node_result(parent_result_obj.dispatch_id, node_result)


def _filter_sublattice_status(
    dispatch_id, node_id, status, node_type, sub_dispatch_id, node_result
):
    if status == Result.COMPLETED and node_type == "sublattice" and not sub_dispatch_id:
        node_result["status"] = RESULT_STATUS.DISPATCHING
    return node_result


async def _make_sublattice_dispatch(dispatch_id: str, node_result: dict):
    try:
        manifest, parent_electron_id = await run_in_executor(
            _make_sublattice_dispatch_helper,
            dispatch_id,
            node_result,
        )

        imported_manifest = await manifest_importer.import_manifest(
            manifest=manifest,
            parent_dispatch_id=dispatch_id,
            parent_electron_id=parent_electron_id,
        )

        return imported_manifest.metadata.dispatch_id

    except ValidationError as ex:
        # Fall back to legacy sublattice handling
        # NB: this loads the JSON sublattice in memory
        json_lattice, parent_electron_id = await run_in_executor(
            _legacy_sublattice_dispatch_helper,
            dispatch_id,
            node_result,
        )
        return await make_dispatch(
            json_lattice,
            dispatch_id,
            parent_electron_id,
        )


def _legacy_sublattice_dispatch_helper(dispatch_id: str, node_result: Dict):
    app_log.debug("falling back to legacy sublattice dispatch")

    result_object = get_result_object(dispatch_id, bare=True)
    node_id = node_result["node_id"]
    parent_node = result_object.lattice.transport_graph.get_node(node_id)
    bg_output = parent_node.get_value("output")

    parent_electron_id = parent_node._electron_id
    json_lattice = bg_output.object_string
    return json_lattice, parent_electron_id


def _make_sublattice_dispatch_helper(dispatch_id: str, node_result: Dict):
    """Helper function for performing DB queries related to sublattices."""
    result_object = get_result_object(dispatch_id, bare=True)
    node_id = node_result["node_id"]
    parent_node = result_object.lattice.transport_graph.get_node(node_id)
    bg_output = parent_node.get_value("output")

    manifest = ResultSchema.parse_raw(bg_output.object_string)
    parent_electron_id = parent_node._electron_id

    return manifest, parent_electron_id


# Common Result object queries


def generate_dispatch_result(
    dispatch_id,
    start_time=None,
    end_time=None,
    status=None,
    error=None,
    result=None,
):
    return {
        "start_time": start_time,
        "end_time": end_time,
        "status": status,
        "error": error,
        "result": result,
    }


# Ensure that a dispatch is only run once; in the future, also check
# if all assets have been uploaded


async def ensure_dispatch(dispatch_id: str) -> bool:
    """Check if a dispatch can be run.

    The following criteria must be met:
    * The dispatch has not been run before.
    * (later) all assets have been uploaded
    """
    return await run_in_executor(
        SRVResult.ensure_run_once,
        dispatch_id,
    )



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Electrons Route"""

import json
import uuid
from typing import List, Optional

from fastapi import APIRouter, HTTPException, Query
from sqlalchemy.orm import Session

import covalent_ui.api.v1.database.config.db as db
from covalent._shared_files.defaults import WAIT_EDGE_NAME
from covalent_dispatcher._core.data_modules import graph as core_graph
from covalent_dispatcher._dal.result import get_result_object
from covalent_ui.api.v1.data_layer.electron_dal import Electrons
from covalent_ui.api.v1.models.electrons_model import (
    ElectronExecutorResponse,
    ElectronFileOutput,
    ElectronFileResponse,
    ElectronResponse,
    Job,
    JobDetails,
    JobDetailsResponse,
    JobsResponse,
)
from covalent_ui.api.v1.utils.file_handle import FileHandler, validate_data
from covalent_ui.api.v1.utils.models_helper import JobsSortBy, SortDirection

routes: APIRouter = APIRouter()


@routes.get("/{dispatch_id}/electron/{electron_id}", response_model=ElectronResponse)
def get_electron_details(dispatch_id: uuid.UUID, electron_id: int):
    """Get Electron Details

    Args:
        electron_id: To fetch electron data with the provided electron id.

    Returns:
        Returns the electron details
    """
    with Session(db.engine) as session:
        electron = Electrons(session)
        result = electron.get_electrons_id(dispatch_id, electron_id)
        if result is None:
            raise HTTPException(
                status_code=400,
                detail=[
                    {
                        "loc": ["path", "dispatch_id"],
                        "msg": f"Dispatch ID {dispatch_id} or Electron ID does not exist",
                        "type": None,
                    }
                ],
            )

        qelectron = {
            "total_quantum_calls": electron.get_total_quantum_calls(
                dispatch_id,
                result["transport_graph_node_id"],
            ),
            "avg_quantum_calls": electron.get_avg_quantum_calls(
                dispatch_id=dispatch_id,
                node_id=result["transport_graph_node_id"],
            ),
        }

        return ElectronResponse(
            id=result["id"],
            node_id=result["transport_graph_node_id"],
            parent_lattice_id=result["parent_lattice_id"],
            type=result["type"],
            storage_path=result["storage_path"],
            name=result["name"],
            status=result["status"],
            started_at=result["started_at"],
            ended_at=result["completed_at"],
            runtime=result["runtime"],
            description="",
            qelectron_data_exists=qelectron["total_quantum_calls"] is not None,
            qelectron=qelectron if qelectron["total_quantum_calls"] is not None else None,
        )


def _get_abstract_task_inputs(dispatch_id: str, node_id: int) -> dict:
    """Return placeholders for the required inputs for a task execution.

    Args:
        dispatch_id: id of the current dispatch
        node_id: Node id of this task in the transport graph.
        node_name: Name of the node.

    Returns: inputs: Input dictionary to be passed to the task with
        `node_id` placeholders for args, kwargs. These are to be
        resolved to their values later.
    """

    abstract_task_input = {"args": [], "kwargs": {}}

    in_edges = core_graph.get_incoming_edges_sync(dispatch_id, node_id)
    for edge in in_edges:
        parent = edge["source"]

        d = edge["attrs"]

        if d["edge_name"] != WAIT_EDGE_NAME:
            if d["param_type"] == "arg":
                abstract_task_input["args"].append((parent, d["arg_index"]))
            elif d["param_type"] == "kwarg":
                key = d["edge_name"]
                abstract_task_input["kwargs"][key] = parent

    sorted_args = sorted(abstract_task_input["args"], key=lambda x: x[1])
    abstract_task_input["args"] = [x[0] for x in sorted_args]

    return abstract_task_input


# Domain: data
def get_electron_inputs(dispatch_id: uuid.UUID, electron_id: int) -> str:
    """
    Get Electron Inputs
    Args:
        dispatch_id: Dispatch id of lattice/sublattice
        electron_id: Transport graph node id of a electron
    Returns:
        Returns the inputs data from Result object
    """

    abstract_inputs = _get_abstract_task_inputs(dispatch_id=str(dispatch_id), node_id=electron_id)

    # Resolve node ids to object strings
    input_assets = {"args": [], "kwargs": {}}

    with Session(db.engine) as session:
        result_object = get_result_object(str(dispatch_id), bare=True)
        tg = result_object.lattice.transport_graph
        for arg in abstract_inputs["args"]:
            node = tg.get_node(node_id=arg, session=session)
            asset = node.get_asset(key="output", session=session)
            input_assets["args"].append(asset)
        for k, v in abstract_inputs["kwargs"].items():
            node = tg.get_node(node_id=v, session=session)
            asset = node.get_asset(key="output", session=session)
            input_assets["kwargs"][k] = asset

    # For now we load the picklefile from the object store into memory, but once
    # TransportableObjects are no longer pickled we will be
    # able to load the byte range for the object string.
    input_args = [asset.load_data() for asset in input_assets["args"]]
    input_kwargs = {k: asset.load_data() for k, asset in input_assets["kwargs"].items()}
    return validate_data({"args": input_args, "kwargs": input_kwargs})


@routes.get("/{dispatch_id}/electron/{electron_id}/details/{name}")
def get_electron_file(dispatch_id: uuid.UUID, electron_id: int, name: ElectronFileOutput):
    """
    Get Electron details
    Args:
        dispatch_id: Dispatch id of lattice/sublattice
        electron_id: Transport graph node id of a electron
        name: refers file type, like inputs, function_string, function, executor, result, value, key,
        stdout, hooks, error, info
    Returns:
        Returns electron details based on the given name
    """

    with Session(db.engine) as session:
        electron = Electrons(session)
        result = electron.get_electrons_id(dispatch_id, electron_id)
        if result is None:
            raise HTTPException(
                status_code=400,
                detail=[
                    {
                        "loc": ["path", "dispatch_id"],
                        "msg": f"Dispatch ID {dispatch_id} or Electron ID does not exist",
                        "type": None,
                    }
                ],
            )
        handler = FileHandler(result["storage_path"])
        if name == "inputs":
            response, python_object = get_electron_inputs(
                dispatch_id=dispatch_id, electron_id=electron_id
            )
            return ElectronFileResponse(data=str(response), python_object=str(python_object))
        elif name == "function_string":
            response = handler.read_from_text(result["function_string_filename"])
            return ElectronFileResponse(data=response)
        elif name == "function":
            response, python_object = handler.read_from_serialized(result["function_filename"])
            return ElectronFileResponse(data=response, python_object=python_object)
        elif name == "executor":
            executor_name = result["executor"]
            executor_data = json.loads(result["executor_data"])
            return ElectronExecutorResponse(
                executor_name=executor_name, executor_details=executor_data
            )
        elif name == "result":
            response, python_object = handler.read_from_serialized(result["results_filename"])
            return ElectronFileResponse(data=str(response), python_object=python_object)
        elif name == "value":
            response = handler.read_from_serialized(result["value_filename"])
            return ElectronFileResponse(data=str(response))
        elif name == "stdout":
            response = handler.read_from_text(result["stdout_filename"])
            return ElectronFileResponse(data=response)
        elif name == "hooks":
            response = handler.read_from_serialized(result["hooks_filename"])
            return ElectronFileResponse(data=response)
        elif name == "error":
            # Error and stderr won't be both populated if `error`
            # is only used for fatal dispatcher-executor interaction errors
            error_response = handler.read_from_text(result["error_filename"])
            stderr_response = handler.read_from_text(result["stderr_filename"])
            if error_response is None:
                error_response = ""
            if stderr_response is None:
                stderr_response = ""
            response = stderr_response + error_response
            return ElectronFileResponse(data=response)
        elif name == "qelectron_db":
            # Since in case of bytes 2 same bytes objects are returned by the handler
            # so we are taking only the first one
            response = handler.read_from_serialized(result["qelectron_db_filename"])[0]
            return ElectronFileResponse(data=response)
        else:
            return ElectronFileResponse(data=None)


@routes.get("/{dispatch_id}/electron/{electron_id}/jobs", response_model=List[Job])
def get_electron_jobs(
    dispatch_id: uuid.UUID,
    electron_id: int,
    sort_by: Optional[JobsSortBy] = JobsSortBy.START_TIME,
    sort_direction: Optional[SortDirection] = SortDirection.DESCENDING,
    count: Optional[int] = None,
    offset: Optional[int] = Query(0),
) -> List[Job]:
    """Get Qelectron Jobs List

    Args:
        dispatch_id: To fetch qelectron data with dispatch id
        electron_id: To fetch qelectron data with the provided electron id.

    Returns:
        Returns the list of qelectron jobs
    """
    with Session(db.engine) as session:
        electron = Electrons(session)
        jobs_response: JobsResponse = electron.get_jobs(
            dispatch_id=dispatch_id,
            electron_id=electron_id,
            sort_by=sort_by,
            sort_direction=sort_direction,
            count=count,
            offset=offset,
        )
        if jobs_response.data is None:
            raise HTTPException(
                status_code=422,
                detail=[{"msg": jobs_response.msg}],
            )
        return jobs_response.data


@routes.get("/{dispatch_id}/electron/{electron_id}/jobs/{job_id}", response_model=JobDetails)
def get_electron_job_overview(dispatch_id: uuid.UUID, electron_id: int, job_id: str) -> JobDetails:
    """Get Electron Job Detail

    Args:
        dispatch_id: To fetch electron data with dispatch id
        electron_id: To fetch electron data with the provided electron id.
        job_id: To fetch appropriate job details with job id

    Returns:
        Returns the electron job details
    """
    with Session(db.engine) as session:
        electron = Electrons(session)
        job_response: JobDetailsResponse = electron.get_job_detail(
            dispatch_id, electron_id, job_id
        )
        if job_response.data is None:
            raise HTTPException(
                status_code=422,
                detail=[{"msg": job_response.msg}],
            )

        return job_response.data



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import codecs
import pickle
import uuid
from datetime import timedelta
from pathlib import Path

from fastapi import HTTPException
from fastapi.responses import JSONResponse
from sqlalchemy import extract, select
from sqlalchemy.sql import func

from covalent._results_manager.results_manager import get_result
from covalent._shared_files import logger
from covalent._shared_files.config import get_config
from covalent_dispatcher._core.execution import _get_task_inputs as get_task_inputs
from covalent_ui.api.v1.data_layer.lattice_dal import Lattices
from covalent_ui.api.v1.database.schema.electron import Electron
from covalent_ui.api.v1.database.schema.lattices import Lattice
from covalent_ui.api.v1.models.electrons_model import JobDetailsResponse, JobsResponse
from covalent_ui.api.v1.utils.file_handle import validate_data
from covalent_ui.api.v1.utils.models_helper import JobsSortBy, SortDirection

app_log = logger.app_log
RESULTS_DIR = Path(get_config("dispatcher")["results_dir"]).resolve()


class Electrons:
    """Electron data access layer"""

    def __init__(self, db_con) -> None:
        self.db_con = db_con

    def electron_exist(self, electron_id: int) -> bool:
        return self.db_con.execute(
            select(Electron).where(Electron.transport_graph_node_id == electron_id)
        ).fetchone()

    def validate_dispatch_and_electron(
        self, dispatch_id: uuid.UUID, electron_id: int, response: JobsResponse
    ) -> (bool, JobsResponse):
        validated = True
        if not Lattices(db_con=self.db_con).dispatch_exist(dispatch_id=dispatch_id):
            validated = False
            response.data = None
            response.msg = f"Dispatch ID {dispatch_id} does not exist"
            return (validated, response)
        if not self.electron_exist(electron_id=electron_id):
            validated = False
            response.data = None
            response.msg = f"Electron ID {electron_id} does not exist"
            return (validated, response)

        return (validated, response)

    def get_jobs(
        self,
        dispatch_id: uuid.UUID,
        electron_id: int,
        sort_by: JobsSortBy,
        sort_direction: SortDirection,
        count,
        offset,
    ) -> JobsResponse:
        try:
            jobs_response = JobsResponse()
            jobs_response.data = None
            jobs_response.msg = None
            (validated, jobs_response) = self.validate_dispatch_and_electron(
                dispatch_id=dispatch_id, electron_id=electron_id, response=jobs_response
            )
            if not validated:
                return jobs_response
            try:
                jobs = self._get_qelectron_db_dict(str(dispatch_id), electron_id)
                if len(jobs) == 0:
                    jobs_response.data = []
                    jobs_response.msg = f"Job details for {dispatch_id} dispatch with {electron_id} node do not exist."
                    return jobs_response
            except Exception as exc:
                app_log.debug(f"Unable to process get jobs \n {exc}")
                jobs_response.data = []
                jobs_response.msg = (
                    f"Jobs for {dispatch_id} dispatch with {electron_id} node do not exist."
                )
                return jobs_response
            jobs_list = list(
                map(
                    lambda circuit: {
                        "job_id": circuit["circuit_id"],
                        "start_time": circuit["save_time"],
                        "executor": circuit["result_metadata"]["executor_name"],
                        "status": "COMPLETED"
                        if len(circuit["result"]) != 0 and len(circuit["result_metadata"]) != 0
                        else "RUNNING",
                    },
                    jobs.values(),
                )
            )
            jobs_list.sort(
                reverse=sort_direction == SortDirection.DESCENDING, key=lambda d: d[sort_by.value]
            )
            result = (
                jobs_list[offset : count + offset] if count is not None else jobs_list[offset:]
            )

            jobs_response.data = result
            return jobs_response
        except Exception as exc:
            app_log.debug(f"Unable to process get jobs \n {exc}")
            jobs_response.msg = "Something went wrong. Please check the log."
            jobs_response.data = None
            return jobs_response

    def get_job_detail(self, dispatch_id, electron_id, job_id) -> JobDetailsResponse:
        try:
            job_detail_response = JobDetailsResponse()
            job_detail_response.data = None
            job_detail_response.msg = None
            (validated, job_detail_response) = self.validate_dispatch_and_electron(
                dispatch_id=dispatch_id, electron_id=electron_id, response=job_detail_response
            )
            if not validated:
                return job_detail_response
            try:
                jobs = self._get_qelectron_db_dict(
                    dispatch_id=str(dispatch_id), node_id=electron_id
                )
                selected_job = jobs[job_id]
            except Exception as exc:
                app_log.debug(f"Unable to process get jobs \n {exc}")
                job_detail_response.data = []
                job_detail_response.msg = (
                    f"Job details for {dispatch_id} dispatch with {electron_id} node do not exist."
                )
                return job_detail_response
            if not selected_job:
                job_detail_response.data = {}
                job_detail_response.msg = (
                    f"Dispatch ID {dispatch_id} or Electron ID does not exist"
                )
                return job_detail_response
            if "result" not in selected_job:
                job_detail_response.data = {}
                job_detail_response.msg = (
                    f"Dispatch ID {dispatch_id} or Electron ID does not exist"
                )
                return job_detail_response
            selected_job["result"] = str(selected_job["result"])[1:-1]
            job_overview = {
                "overview": {
                    "job_name": selected_job["circuit_name"]
                    if "circuit_name" in selected_job
                    else None,
                    "backend": selected_job["result_metadata"]["executor_backend_name"]
                    if "result_metadata" in selected_job
                    and "executor_backend_name" in selected_job["result_metadata"]
                    else None,
                    "time_elapsed": selected_job["execution_time"]
                    if "execution_time" in selected_job
                    else None,
                    "result": selected_job["result"] if "result" in selected_job else None,
                    "status": "COMPLETED"
                    if len(selected_job["result"]) != 0
                    and len(selected_job["result_metadata"]) != 0
                    else "RUNNING",
                    "start_time": selected_job["save_time"]
                    if "save_time" in selected_job
                    else None,
                },
                "circuit": {
                    "total_qbits": None,
                    "depth": None,
                    "circuit_diagram": selected_job["circuit_diagram"]
                    if "circuit_diagram" in selected_job
                    else None,
                },
                "executor": {
                    "name": selected_job["qexecutor"]["name"]
                    if "qexecutor" in selected_job and "name" in selected_job["qexecutor"]
                    else None,
                    "executor": str(selected_job["qexecutor"])
                    if "qexecutor" in selected_job
                    else None,
                },
            }

            job_overview["overview"]["end_time"] = (
                selected_job["save_time"] + timedelta(seconds=selected_job["execution_time"])
                if job_overview["overview"]["start_time"]
                and job_overview["overview"]["time_elapsed"]
                else None
            )
            if selected_job["qnode_specs"] is not None:
                job_overview["circuit"]["total_qbits"] = (
                    selected_job["qnode_specs"]["num_used_wires"]
                    if "num_used_wires" in selected_job["qnode_specs"]
                    else None
                )
                job_overview["circuit"]["depth"] = (
                    selected_job["qnode_specs"]["depth"]
                    if "depth" in selected_job["qnode_specs"]
                    else None
                )
                gate_sizes = (
                    selected_job["qnode_specs"]["gate_sizes"]
                    if "gate_sizes" in selected_job["qnode_specs"]
                    else None
                )
                if gate_sizes:
                    for gate_wires, gate_count in gate_sizes.items():
                        job_overview["circuit"][f"qbit{gate_wires}_gates"] = gate_count

            job_detail_response.data = job_overview
            job_detail_response.msg = ""
            return job_detail_response
        except Exception as exc:
            app_log.debug(f"Unable to process get job details \n {exc}")
            job_detail_response.msg = "Something went wrong. Please check the log."
            job_detail_response.data = None
            return job_detail_response

    def get_electrons_id(self, dispatch_id, electron_id) -> Electron:
        """
        Read electron table by electron id
        Args:
            electron_id: Refers to the electron's PK
        Return:
            Electron with PK as electron_id
        """
        data = (
            self.db_con.query(
                Electron.id,
                Electron.transport_graph_node_id,
                Electron.parent_lattice_id,
                Electron.type,
                Electron.storage_path,
                Electron.function_filename,
                Electron.function_string_filename,
                Electron.executor,
                Electron.executor_data,
                Electron.results_filename,
                Electron.value_filename,
                Electron.stdout_filename,
                Electron.hooks_filename,
                Electron.stderr_filename,
                Electron.error_filename,
                Electron.name,
                Electron.status,
                Electron.job_id,
                Electron.qelectron_data_exists,
                Electron.started_at.label("started_at"),
                Electron.completed_at.label("completed_at"),
                (
                    (
                        func.coalesce(
                            extract("epoch", Electron.completed_at),
                            extract("epoch", func.now()),
                        )
                        - extract("epoch", Electron.started_at)
                    )
                    * 1000
                ).label("runtime"),
            )
            .join(Lattice, Lattice.id == Electron.parent_lattice_id)
            .filter(
                Lattice.dispatch_id == str(dispatch_id),
                Electron.transport_graph_node_id == electron_id,
            )
            .first()
        )
        return data

    def get_total_quantum_calls(self, dispatch_id, node_id):
        qdb = self._get_qelectron_db_dict(dispatch_id=str(dispatch_id), node_id=node_id)
        return None if len(qdb) == 0 else len(qdb)

    def get_avg_quantum_calls(self, dispatch_id, node_id):
        jobs = self._get_qelectron_db_dict(dispatch_id=str(dispatch_id), node_id=node_id)
        if len(jobs) == 0:
            return None

        time = [jobs[value]["execution_time"] for value in jobs]
        return sum(time) / len(time)

    def get_electron_inputs(self, dispatch_id: uuid.UUID, electron_id: int) -> str:
        """
        Get Electron Inputs
        Args:
            dispatch_id: Dispatch id of lattice/sublattice
            electron_id: Transport graph node id of a electron
        Returns:
            Returns the inputs data from Result object
        """

        result = get_result(dispatch_id=str(dispatch_id), wait=False)
        if isinstance(result, JSONResponse) and result.status_code == 404:
            raise HTTPException(status_code=400, detail=result)
        result_object = pickle.loads(codecs.decode(result["result"].encode(), "base64"))
        electron_result = self.get_electrons_id(dispatch_id, electron_id)
        inputs = get_task_inputs(
            node_id=electron_id, node_name=electron_result.name, result_object=result_object
        )
        return validate_data(inputs)

    def _get_qelectron_db_dict(self, dispatch_id: str, node_id: int) -> dict:
        """Return the QElectron DB for a given node."""

        try:
            from covalent.quantum.qserver.database import Database

            electron = self.get_electrons_id(dispatch_id, node_id)

            database = Database(electron.storage_path)
            qelectron_db_dict = database.get_db_dict(
                dispatch_id=dispatch_id, node_id=node_id, direct_path=True
            )

            return qelectron_db_dict
        except ImportError:
            app_log.debug("QElectron not installed.")
            return {}



# Copyright 2023 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""""
Pennylane-Qiskit device that uses the local Qiskit `Sampler` primitive
"""

from typing import Any, List, Tuple

from devices_base import QiskitSamplerDevice
from qiskit.primitives import Sampler as LocalSampler


class QiskitLocalSampler(QiskitSamplerDevice):
    """
    Pennylane device that runs circuits using the local `qiskit.primitives.Sampler`
    """

    short_name = "local_sampler"

    def __init__(self, wires: int, shots: int, **_):
        self.circuit = None
        self.transpile_args = {}

        super().__init__(
            wires=wires,
            shots=shots,
            backend_name="None",
            local_transpile=False,
            transpile_backend=None,
            service_init_kwargs={},
        )

        self._metadatas = []

    def batch_execute(self, circuits, timeout: int = None):
        # pylint: disable=missing-function-docstring

        n_original_circuits = len(circuits)
        circuits = self.broadcast_tapes(circuits)
        n_circuits = len(circuits)

        # Create circuit objects and apply diagonalizing gates
        compiled_circuits = self.compile_circuits(circuits)

        # Send the batch of circuit objects using sampler.run
        sampler = LocalSampler()
        job = sampler.run(compiled_circuits)
        job_result = job.result()  # local operation, no API call here

        # Increment counter for number of executions of qubit device
        self._num_executions += 1

        # Compute statistics using the state and/or samples
        results = []
        self._metadatas = []

        for circuit, quasi_dist in zip(circuits, job_result.quasi_dists):
            # Process quasi-distribution into expected numerical result
            res = self._process_batch_execute_result(circuit, quasi_dist)
            results.append(res)

            # Construct metadata
            metadata = {
                "result_object": job_result,
                "num_measurements": self._num_executions,
            }
            self._metadatas.append(metadata)

        # Update tracker
        if self.tracker.active:
            self.tracker.update(batches=1, batch_len=n_circuits)
            self.tracker.record()

        if n_original_circuits != n_circuits:
            self._n_circuits = n_circuits
            self._n_original_circuits = n_original_circuits
            results = self._vector_results(results)

        return results

    def post_process(self, *args) -> Tuple[Any, List[dict]]:
        results = args[1]
        metadatas = self._metadatas.copy()
        self._metadatas = None
        return results, metadatas

    def post_process_all(self, *args) -> Tuple[Any, List[dict]]:
        # Dummy method for local Sampler, redirects to `post_process`
        return self.post_process(*args)



# Copyright 2023 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Wrappers for the existing Pennylane-Qiskit interface"""
import warnings
from abc import ABC, abstractmethod
from contextlib import contextmanager
from math import sqrt
from typing import Any, List, Sequence, Tuple, Union

import numpy as np
import pennylane.numpy as pnp
from pennylane.transforms import broadcast_expand, map_batch_transform
from pennylane_qiskit.qiskit_device import QiskitDevice
from qiskit.compiler import transpile
from sessions import init_runtime_service


class _PennylaneQiskitDevice(QiskitDevice, ABC):
    # pylint: disable=too-many-instance-attributes
    """
    A replacement for Pennylane's `QiskitDevice` that uses `QiskitRuntimeService`
    instead of the older `qiskit.providers.provider.ProviderV1`
    """

    name = "Custom Plugin for Pennylane Circuits on Qiskit IBM Runtime"
    version = "0.0.1"
    author = "AQ"

    def __init__(
        self,
        wires: int,
        shots: Union[None, int, Sequence[int], Sequence[Union[int, Sequence[int]]]],
        backend_name: str,
        local_transpile: bool,
        service_init_kwargs: dict,
        **kwargs,
    ):
        # proxy for client-side `pennylane.active_return()` status
        self.pennylane_active_return = True

        super(QiskitDevice, self).__init__(wires=wires, shots=shots)

        self.shots = shots
        self.backend_name = backend_name
        self.local_transpile = local_transpile
        self.service_init_kwargs = service_init_kwargs
        self.device_kwargs = kwargs

        self._service = None
        self._backend = None

        self.reset()
        self.process_kwargs(kwargs)

        self._state = None

    @classmethod
    def capabilities(cls):
        capabilities = super().capabilities().copy()
        capabilities.update(supports_broadcasting=True)
        return capabilities

    @property
    def service(self):
        """
        Lazy service property - enables serialization
        """
        if self._service is None:
            # assign cached service instance
            self._service = init_runtime_service(**self.service_init_kwargs)
        return self._service

    @property
    def backend(self):
        """
        Lazy backend property - enables serialization
        """
        if self._backend is None:
            self._backend = self.service.get_backend(self.backend_name)
        return self._backend

    def process_kwargs(self, kwargs):
        """
        Override original method from `pennylane_qiskit.qiskit_device.QiskitDevice`
        to avoid accessing backend property, unnecessary init of service
        """
        self.compile_backend = None
        self.set_transpile_args(**kwargs)
        self.run_args = kwargs.copy()

    def map_wires(self, wires):
        return wires

    def broadcast_tapes(self, tapes):
        """
        Broadcast tapes for batch execution (if necessary)
        """
        if all(tape.batch_size for tape in tapes):
            expanded_tapes, _ = map_batch_transform(broadcast_expand, tapes)
            return expanded_tapes

        return tapes

    @abstractmethod
    def post_process(self, *args) -> Tuple[Any, List[dict]]:
        """
        Obtain metadata; make blocking API call to Qiskit Runtime.
        """
        raise NotImplementedError

    @abstractmethod
    def post_process_all(self, *args) -> Tuple[Any, List[dict]]:
        """
        Single job version of `post_process`.
        """
        raise NotImplementedError


class QiskitSamplerDevice(_PennylaneQiskitDevice):

    """
    A base class for devices that use the Sampler primitive.
    """

    default_shots = 1024

    def __init__(
        self,
        wires: int,
        shots: Union[None, int, Sequence[int], Sequence[Union[int, Sequence[int]]]],
        backend_name: str,
        local_transpile: bool,
        service_init_kwargs: dict,
        **kwargs,
    ):
        super().__init__(
            wires=wires,
            shots=shots,
            backend_name=backend_name,
            local_transpile=local_transpile,
            service_init_kwargs=service_init_kwargs,
            **kwargs,
        )

        self._default_shots = QiskitSamplerDevice.default_shots

        if self.shots is None:
            warnings.warn(
                "The Qiskit Sampler device does not support analytic expectation values. "
                "The number of shots can not be None. "
                f"{self._default_shots} shots will be used."
            )

        self._current_quasi_dist = None
        self._dummy_state = None

        # Used for reshaping results.
        self._n_original_circuits = None
        self._n_circuits = None

    def compile_circuits(self, circuits):
        """
        Override `QiskitDevice.compile_circuits` to include `unwrap` context.
        This step avoids potential parameter type errors during compilation.
        """
        compiled_circuits = []
        for circuit in circuits:
            # Unwraps a quantum script with tensor-like parameters to numpy arrays.
            with circuit.unwrap():
                qiskit_circuit = super().compile_circuits([circuit]).pop()
                compiled_circuits.append(qiskit_circuit)

        return compiled_circuits

    def compile(self):
        """
        Overrides `QiskitDevice.compile` with custom choice logic for the `backend`
        argument during transpilation.
        """
        backend = self.backend if self.local_transpile else None
        return transpile(self._circuit, backend=backend, **self.transpile_args)

    @property
    def asarray(self):
        """
        Array function property to return Pennylane tensors instead of NumPy arrays.
        """
        if self._asarray is np.asarray:
            return pnp.asarray
        return self._asarray

    @property
    def _state(self):
        """
        Override `self._state` to avoid unnecessary state reconstruction for
        every execution result.
        """
        if self._dummy_state is None:
            return self.dist_get_state()
        return self._dummy_state

    @property
    def vector_shape(self):
        """
        Used to reshape results in case of vector inputs.
        """
        n = self._n_original_circuits
        m = self._n_circuits // n
        return (n, m)

    @_state.setter
    def _state(self, state):
        """
        Allows `self._state` to be set as if instance attribute
        """
        self._dummy_state = state

    def dist_get_state(self):
        """
        Generate a state-vector from a quasi-distribution
        """
        N = 2 ** len(self.wires)

        state = np.zeros(N)
        for i in range(N):
            # invert bit string representation of i
            probs_idx = int(bin(i)[2:].zfill(len(self.wires))[::-1], base=2)

            if prob := self._current_quasi_dist.get(i):
                state[probs_idx] = sqrt(prob)
            else:
                state[probs_idx] = 0

        return self._asarray(state, dtype=self.C_DTYPE)

    def dist_generate_samples(self, quasi_dist):
        """
        Generate samples from a quasi-distribution
        """

        dist_bin = quasi_dist.binary_probabilities()
        dist_bin = self.normalize_probability(dist_bin)

        bit_strings = list(dist_bin)
        probs = [dist_bin[bs] for bs in bit_strings]

        # generate artificial samples from quasi-distribution probabilities
        size = self.shots if self.shots else self._default_shots
        bit_samples = np.random.choice(bit_strings, size=size, p=probs)
        return np.vstack([np.array([int(i) for i in s[::-1]]) for s in bit_samples])

    @contextmanager
    def set_distribution(self, quasi_dist):
        """
        Set the current quasi-distribution for statistics computations.
        """
        self._current_quasi_dist = quasi_dist
        try:
            yield
        finally:
            self._current_quasi_dist = None

    @staticmethod
    def normalize_probability(distribution: dict) -> dict:
        """
        Some IBMQ backends may return small NEGATIVE quasi-probabilities
        instead of ~0, when using `Sampler`.

        Below is an example of this, as observed with "ibmq_lima":

        # quasi-probabilities list
        [
            ...
            {
                "10": 0.0026338769545423205,
                "11": 0.5150691505205544,
                "00": 0.4867495783187873,
                "01": -0.004452605793883964
            },
            ...
        ]

        To avoid errors due to negative probability, this function zeros any
        negative values, then re-normalizes the quasi probabilities.
        """

        dist = distribution.copy()

        # Negative probabilities are presumed to be zero.
        delete_keys = []
        for bit_string, quasi_prob in dist.items():
            if quasi_prob <= 0:
                delete_keys.append(bit_string)

        # Zero probabilities should be omitted from the dict.
        for key in delete_keys:
            del dist[key]

        if len(dist) == 0:
            # This should never happen.
            raise RuntimeError(f"No positive probabilities exist in {distribution}.")

        # Re-normalize the remaining quasi-probabilities.
        total_prob = sum(dist.values())
        for bit_string, quasi_prob in dist.items():
            dist[bit_string] = quasi_prob / total_prob

        return dist

    def _process_batch_execute_result(self, circuit, quasi_dist) -> Any:
        # Update the tracker
        if self.tracker.active:
            self.tracker.update(executions=1, shots=self.shots)
            self.tracker.record()

        # Generate computational basis samples
        self._samples = self.dist_generate_samples(quasi_dist)

        with self.set_distribution(quasi_dist):
            if not self.pennylane_active_return:
                res = self._statistics_legacy(circuit)
                return self.asarray(res)

            if self._shot_vector is not None:
                res = self.shot_vec_statistics(circuit)
            else:
                res = self.statistics(circuit)

        if len(circuit.measurements) > 1:
            return tuple(res)

        return self.asarray(res[0])

    def _vector_results(self, res):
        """
        Process the result of a vectorized QElectron call.
        """

        res = pnp.asarray(res)

        if self.pennylane_active_return:
            return [res] if res.ndim > 1 else list(res.reshape(self.vector_shape))

        res = res.reshape(-1)
        return [res] if res.ndim > 1 else [res.reshape(self.vector_shape)]



# Copyright 2023 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional

import pennylane as qml
from braket.aws import AwsQuantumTask, AwsQuantumTaskBatch
from pydantic import Field

from covalent._shared_files.config import get_config
from covalent.executor.qbase import BaseThreadPoolQExecutor, QCResult, get_thread_pool

__all__ = ["BraketQubitExecutor", "LocalBraketQubitExecutor"]

_QEXECUTOR_PLUGIN_DEFAULTS = {
    "BraketQubitExecutor": {
        "device_arn": "",
        "poll_timeout_seconds": AwsQuantumTask.DEFAULT_RESULTS_POLL_TIMEOUT,
        "poll_interval_seconds": AwsQuantumTask.DEFAULT_RESULTS_POLL_INTERVAL,
        "max_connections": AwsQuantumTaskBatch.MAX_CONNECTIONS_DEFAULT,
        "max_retries": AwsQuantumTaskBatch.MAX_RETRIES,
    },
    "LocalBraketQubitExecutor": {
        "backend": "default",
    },
}


class BraketQubitExecutor(BaseThreadPoolQExecutor):
    """
    The remote Braket executor based on the existing Pennylane Braket
    qubit device. Usage of this device requires valid AWS credentials as
    set up following the instructions at
    https://github.com/aws/amazon-braket-sdk-python#prerequisites.

    Attributes:
        max_jobs:
            maximum number of parallel jobs sent by threads on :code:`batch_submit`.
        shots: number of shots used to estimate quantum observables.
        device_arn:
            an alpha-numeric code (arn=Amazon Resource Name) specifying a quantum device.
        poll_timeout_seconds:
            number of seconds before a poll to remote device is considered timed-out.
        poll_interval_seconds:
            number of seconds between polling of a remote device's status.
        aws_session:
            An :code:`AwsSession` object created to manage interactions with AWS services,
            to be supplied if extra control is desired.
        parallel: turn parallel execution on or off.
        max_parallel: the maximum number of circuits to be executed in parallel.
        max_connections: the maximum number of connections in the :code:`Boto3` connection pool.
        max_retries: the maximum number of time a job will be re-sent if it failed
        s3_destination_folder: Name of the S3 bucket and folder, specified as a tuple.
        run_kwargs: Variable length keyword arguments for :code:`braket.devices.Device.run()`

    """

    device_arn: str = Field(
        default_factory=lambda: get_config("qelectron")["BraketQubitExecutor"]["device_arn"]
    )
    poll_timeout_seconds: float = Field(
        default_factory=lambda: get_config("qelectron")["BraketQubitExecutor"][
            "poll_timeout_seconds"
        ]
    )
    poll_interval_seconds: float = Field(
        default_factory=lambda: get_config("qelectron")["BraketQubitExecutor"][
            "poll_interval_seconds"
        ]
    )
    max_connections: int = Field(
        default_factory=lambda: get_config("qelectron")["BraketQubitExecutor"]["max_connections"]
    )
    max_retries: int = Field(
        default_factory=lambda: get_config("qelectron")["BraketQubitExecutor"]["max_retries"]
    )
    max_jobs: int = 20
    aws_session: Optional[str] = None  # not actually a str. Fix.
    parallel: bool = False
    max_parallel: Optional[int] = None
    s3_destination_folder: tuple = ()
    run_kwargs: dict = {}

    def batch_submit(self, qscripts_list):
        """
        Submit qscripts for execution using :code:`max_jobs`-many threads.

        Args:
            qscripts_list: a list of Pennylane style :code:`QuantumScripts`

        Returns:
            jobs: a :code:`list` of tasks subitted by threads.
        """

        p = get_thread_pool(self.max_jobs)
        jobs = []
        for qscript in qscripts_list:
            dev = qml.device(
                "braket.aws.qubit",
                wires=self.qelectron_info.device_wires,
                device_arn=self.device_arn,
                s3_destination_folder=self.s3_destination_folder,
                shots=self.override_shots,
                poll_timeout_seconds=self.poll_timeout_seconds,
                poll_interval_seconds=self.poll_interval_seconds,
                aws_session=self.aws_session,
                parallel=self.parallel,
                max_parallel=self.max_parallel,
                max_connections=self.max_connections,
                max_retries=self.max_retries,
                **self.run_kwargs,
            )

            result_obj = QCResult.with_metadata(
                device_name=dev.short_name,
                executor=self,
            )

            jobs.append(p.submit(self.run_circuit, qscript, dev, result_obj))

        return jobs

    def dict(self, *args, **kwargs):
        dict_ = super().dict(*args, **kwargs)
        # needed to make the dict method hashable and jsonable
        dict_["run_kwargs"] = tuple(dict_["run_kwargs"].items())
        return dict_


class LocalBraketQubitExecutor(BaseThreadPoolQExecutor):
    """
    The local Braket executor based on the existing Pennylane local Braket qubit device.

    Attributes:
        max_jobs: maximum number of parallel jobs sent by processes on :code:`batch_submit`.
        shots: number of shots used to estimate quantum observables.
        backend:
            The name of the simulator backend. Defaults to the :code:`"default"`
            simulator backend name.
        run_kwargs: Variable length keyword arguments for :code:`braket.devices.Device.run()`.
    """

    backend: str = Field(
        default_factory=lambda: get_config("qelectron")["LocalBraketQubitExecutor"]["backend"]
    )
    max_jobs: int = 20
    run_kwargs: dict = {}

    def batch_submit(self, qscripts_list):
        """
        Submit qscripts for execution using :code:`num_threads`-many threads.

        Args:
            qscripts_list: a list of Pennylane style :code:`QuantumScripts`.

        Returns:
            jobs: a :code:`list` of tasks subitted by threads.
        """

        p = get_thread_pool(self.max_jobs)
        jobs = []
        for qscript in qscripts_list:
            dev = qml.device(
                "braket.local.qubit",
                wires=self.qelectron_info.device_wires,
                backend=self.backend,
                shots=self.override_shots,
                **self.run_kwargs,
            )

            result_obj = QCResult.with_metadata(
                device_name=dev.short_name,
                executor=self,
            )

            jobs.append(p.submit(self.run_circuit, qscript, dev, result_obj))

        return jobs

    def dict(self, *args, **kwargs):
        dict_ = super().dict(*args, **kwargs)
        # needed to make the dict method hashable and jsonable
        dict_["run_kwargs"] = tuple(dict_["run_kwargs"].items())
        return dict_



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""General utils for Covalent."""

import inspect
import shutil
import socket
from datetime import timedelta
from typing import Callable, Dict, List, Tuple

from . import logger
from .config import get_config

app_log = logger.app_log
log_stack_info = logger.log_stack_info

DEFAULT_UI_ADDRESS = get_config("user_interface.address")
DEFAULT_UI_PORT = get_config("user_interface.port")


def get_ui_url(path):
    baseUrl = f"http://{DEFAULT_UI_ADDRESS}:{DEFAULT_UI_PORT}"
    return f"{baseUrl}{path}"


def get_random_available_port() -> int:
    """
    Return a random port that is available on the machine
    """
    sock = socket.socket()
    sock.bind(("127.0.0.1", 0))
    port = sock.getsockname()[1]
    sock.close()
    return port


def get_timedelta(time_limit: str) -> timedelta:
    """
    Get timedelta from a compatible time limit string passed to the lattice/electron decorator.

    Args:
        time_limit: The time limit string.
    Returns:
        timedelta: The `datetime.timedelta` object.
    """

    days, hours, minutes, seconds = time_limit.split("-")[0], *time_limit.split("-")[1].split(":")
    return timedelta(
        days=int(days),
        hours=int(hours),
        minutes=int(minutes),
        seconds=int(seconds),
    )


def reformat(t: int) -> str:
    """
    Reformat an integer to a readable time-like string. For example, if t = 1, return "01".

     Args:
         t: The integer to reformat.

     Returns:
         ref_string: The reformatted string.
    """

    return f"0{t}" if len(str(t)) == 1 else str(t)


def get_time(time_delta: timedelta) -> str:
    """
    Get a compatible time string from a timedelta object.

    Args:
        time_delta: The timedelta object.
    Returns:
        time_string: The compatible reformatted time string.
    """

    days = reformat(time_delta.days)
    hours = reformat(time_delta.seconds // 3600)
    minutes = reformat((time_delta.seconds // 60) % 60)
    seconds = reformat(time_delta.seconds % 60)
    return f"{days}-{hours}:{minutes}:{seconds}"


def filter_null_metadata(meta_dict: dict) -> Dict:
    """Filter out metadata that is None or empty."""
    return {k: v for k, v in meta_dict.items() if v}


def get_serialized_function_str(function):
    """
    Generates a string representation of a function definition
    including the decorators on it.

    Args:
        function: The function whose definition is to be convert to a string.

    Returns:
        function_str: The string representation of the function definition.
    """

    input_function = function
    # If a Lattice or electron object was passed as the function input, we need the
    # (deserialized) underlying function describing the lattice.
    while hasattr(input_function, "workflow_function"):
        input_function = input_function.workflow_function.get_deserialized()

    try:
        # function_str is the string representation of one function, with decorators, if any.
        function_str = inspect.getsource(input_function)
    except Exception:
        function_str = f"# {function.__name__} was not inspectable"

    return function_str + "\n\n"


def get_imports(func: Callable) -> Tuple[str, List[str]]:
    """
    Given an input workflow function, find the imports that were used, and determine
        which ones are Covalent-related.

    Args:
        func: workflow function.

    Returns:
        A tuple consisting of a string of import statements and a set of names that
            Covalent-related modules have been imported as.
    """

    imports_str = ""
    cova_imports = []
    for i, j in func.__globals__.items():
        if inspect.ismodule(j) or (
            inspect.isfunction(j) and j.__name__ in ["lattice", "electron"]
        ):
            if j.__name__ == i:
                import_line = f"import {j.__name__}\n"
            else:
                import_line = f"import {j.__name__} as {i}\n"

            if j.__name__ in ["covalent", "lattice", "electron"]:
                import_line = f"# {import_line}"
                cova_imports.append(i)

            imports_str += import_line

    return imports_str, cova_imports


def required_params_passed(func: Callable, kwargs: Dict) -> bool:
    """
    DEPRECATED: Check to see that values for all parameters without default values have been passed.

    Args:
        func: Callable function.
        kwargs: Parameter list with passed values.

    Returns:
        status: Whether all the parameters required for the callable function has been passed.
    """

    required_arg_set = set({})
    sig = inspect.signature(func)
    for param in sig.parameters.values():
        if param.default is param.empty:
            required_arg_set.add(str(param))

    return required_arg_set.issubset(set(kwargs.keys()))


def get_named_params(func, args, kwargs):
    ordered_params_dict = inspect.signature(func).parameters

    named_args = {}
    named_kwargs = {}

    for ind, parameter_dict in enumerate(ordered_params_dict.items()):
        param_name, param = parameter_dict

        if param.kind in [param.POSITIONAL_ONLY, param.POSITIONAL_OR_KEYWORD]:
            if param_name in kwargs:
                named_kwargs[param_name] = kwargs[param_name]
            elif ind < len(args):
                named_args[param_name] = args[ind]
        elif param.kind == param.VAR_POSITIONAL:
            for i in range(ind, len(args)):
                named_args[f"arg[{i}]"] = args[i]
        elif param.kind == param.VAR_KEYWORD:
            for key, value in kwargs.items():
                if key != param_name:
                    named_kwargs[key] = value
        elif param.kind == param.KEYWORD_ONLY:
            if param_name in kwargs:
                named_kwargs[param_name] = kwargs[param_name]

    if len(args) > len(named_args):
        raise ValueError(
            f"Too many positional arguments given, expected {len(named_args)}, received {len(args)}"
        )

    if len(kwargs) > len(named_kwargs):
        extra_supplied_kwargs = ", ".join(sorted(set(kwargs.keys()) - set(named_kwargs.keys())))
        raise ValueError(f"Unexpected keyword arguments: {extra_supplied_kwargs}")

    return (named_args, named_kwargs)


def format_server_url(hostname: str = None, port: int = None) -> str:
    if hostname is None:
        hostname = get_config("dispatcher.address")
    if port is None:
        port = int(get_config("dispatcher.port"))

    url = hostname
    if not url.startswith("http"):
        url = f"https://{url}" if port == 443 else f"http://{url}"

    # Inject port
    if port not in [80, 443]:
        parts = url.split("/")
        url = "".join(["/".join(parts[:3])] + [f":{port}/"] + ["/".join(parts[3:])])

    return url.strip("/")


# For use by LocalDispatcher and ResultsManager when running Covalent
# server locally
def copy_file_locally(src_uri, dest_uri):
    scheme_prefix = "file://"
    if src_uri.startswith(scheme_prefix):
        src_path = src_uri[len(scheme_prefix) :]
    else:
        raise TypeError(f"{src_uri} is not a valid URI")
        # src_path = src_uri
    if dest_uri.startswith(scheme_prefix):
        dest_path = dest_uri[len(scheme_prefix) :]
    else:
        raise TypeError(f"{dest_uri} is not a valid URI")

    shutil.copyfile(src_path, dest_path)


def get_qelectron_db_path(dispatch_id: str, task_id: int):
    """
    Return the path to the Qelectron database for a given dispatch_id and task_id.

    This is a proxy to qelectron_utils.get_qelectron_db_path() for removing qelectron dependency.
    """

    try:
        from .qelectron_utils import get_qelectron_db_path

        return get_qelectron_db_path(dispatch_id, task_id)
    except ImportError:
        return None



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""FastAPI models for /api/v1/resultv2 endpoints"""

from typing import Dict, Optional

from pydantic import BaseModel, field_validator

from .asset import AssetSchema
from .transport_graph import TransportGraphSchema

LATTICE_METADATA_KEYS = {
    "__name__",
    "python_version",
    "covalent_version",
    # user dependent metadata
    "executor",
    "workflow_executor",
    "executor_data",
    "workflow_executor_data",
}

LATTICE_ASSET_KEYS = {
    "workflow_function",
    "workflow_function_string",
    "__doc__",
    "inputs",
    "named_args",
    "named_kwargs",
    "cova_imports",
    "lattice_imports",
    # user dependent assets
    "hooks",
}

LATTICE_FUNCTION_FILENAME = "function.tobj"
LATTICE_FUNCTION_STRING_FILENAME = "function_string.txt"
LATTICE_DOCSTRING_FILENAME = "function_docstring.txt"
LATTICE_EXECUTOR_DATA_FILENAME = "executor_data.pkl"
LATTICE_WORKFLOW_EXECUTOR_DATA_FILENAME = "workflow_executor_data.pkl"
LATTICE_ERROR_FILENAME = "error.log"
LATTICE_INPUTS_FILENAME = "inputs.tobj"
LATTICE_NAMED_ARGS_FILENAME = "named_args.tobj"
LATTICE_NAMED_KWARGS_FILENAME = "named_kwargs.tobj"
LATTICE_RESULTS_FILENAME = "results.tobj"
LATTICE_HOOKS_FILENAME = "hooks.json"
LATTICE_CALL_BEFORE_FILENAME = "call_before.json"
LATTICE_CALL_AFTER_FILENAME = "call_after.json"
LATTICE_COVA_IMPORTS_FILENAME = "cova_imports.json"
LATTICE_LATTICE_IMPORTS_FILENAME = "lattice_imports.txt"
LATTICE_STORAGE_TYPE = "file"


ASSET_FILENAME_MAP = {
    "workflow_function": LATTICE_FUNCTION_FILENAME,
    "workflow_function_string": LATTICE_FUNCTION_STRING_FILENAME,
    "doc": LATTICE_DOCSTRING_FILENAME,
    "inputs": LATTICE_INPUTS_FILENAME,
    "named_args": LATTICE_NAMED_ARGS_FILENAME,
    "named_kwargs": LATTICE_NAMED_KWARGS_FILENAME,
    "cova_imports": LATTICE_COVA_IMPORTS_FILENAME,
    "lattice_imports": LATTICE_LATTICE_IMPORTS_FILENAME,
    "hooks": LATTICE_HOOKS_FILENAME,
}


class LatticeAssets(BaseModel):
    workflow_function: AssetSchema
    workflow_function_string: AssetSchema
    doc: AssetSchema  # __doc__
    inputs: AssetSchema
    named_args: AssetSchema
    named_kwargs: AssetSchema
    cova_imports: AssetSchema
    lattice_imports: AssetSchema

    # lattice.metadata
    hooks: AssetSchema


class LatticeMetadata(BaseModel):
    name: str  # __name__
    executor: str
    executor_data: dict
    workflow_executor: str
    workflow_executor_data: dict
    python_version: Optional[str] = None
    covalent_version: Optional[str] = None


class LatticeSchema(BaseModel):
    metadata: LatticeMetadata
    assets: LatticeAssets
    custom_assets: Optional[Dict[str, AssetSchema]] = None

    transport_graph: TransportGraphSchema

    @field_validator("custom_assets")
    def check_custom_asset_keys(cls, v):
        if v is not None:
            for key in v:
                if key in ASSET_FILENAME_MAP:
                    raise ValueError(f"Asset {key} conflicts with built-in key")
        return v



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod
from typing import Any

from .._results_manager.result import Result
from .._workflow.lattice import Lattice


class BaseDispatcher(ABC):
    """
    The base dispatcher class to allow defining custom dispatchers.
    Subclassed dispatchers must implement the `dispatch` method and
    `dispatch_sync` methods. These dispatchers act as the interface
    to the covalent dispatcher server and using the `dispatch` method
    will allow the user to submit jobs to the dispatcher server.
    """

    def __init__(self) -> None:
        pass

    @abstractmethod
    def dispatch(self, lattice: Lattice, custom_variables: Any) -> None:
        """
        Dispatch a lattice to the dispatcher server.
        """

        pass

    @abstractmethod
    def dispatch_sync(self, lattice: Lattice, custom_variables: Any) -> Result:
        """
        Dispatch a lattice to the dispatcher server and wait for the results.
        """

        pass



# Copyright 2023 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=invalid-name

"""Hopefully temporary custom tools to handle pickling and/or un-pickling."""

from contextlib import contextmanager
from typing import Any, Callable, Tuple

from pennylane.ops.qubit.observables import Projector

_PENNYLANE_METHOD_OVERRIDES = (
    # class, method_name, method_func
    (Projector, "__reduce__", lambda self: (Projector, (self.data[0], self.wires))),
)


def _qml_mods_pickle(func: Callable) -> Callable:
    """
    A decorator that applies overrides to select PennyLane objects, making them
    pickleable and/or un-pickleable in the local scope.
    """

    def _wrapper(*args, **kwargs):
        with _method_overrides(_PENNYLANE_METHOD_OVERRIDES):
            return func(*args, **kwargs)

    return _wrapper


@contextmanager
def _method_overrides(overrides: Tuple[Any, str, Callable]) -> None:
    """
    Creates a context where all `overrides` are applied on entry and un-applied on exit.
    """

    unapply_overrides = None
    try:
        unapply_overrides = _apply_method_overrides(overrides)
        yield
    finally:
        unapply_overrides()


def _apply_method_overrides(overrides: Tuple[Any, str, Callable]) -> Callable:
    """
    This function is called by the `_method_overrides()` context manager.

    It applies the overrides in `_METHOD_OVERRIDES` to the corresponding objects
    and returns a function that can later restore those objects.
    """

    restoration_list = []
    for cls, method_name, func in overrides:
        # Attribute will be deleted later if `attr` is a length-1 tuple.
        attr = (method_name,)
        if hasattr(cls, method_name):
            # Attribute will be restored later to the corresponding method.
            attr += (getattr(cls, method_name),)

        # Store attribute information.
        restoration_list.append(attr)

        # Use `func` to create or replace the method by name.
        setattr(cls, method_name, func)

    def _unapply_overrides():
        for attr in restoration_list:
            # Here `attr` is `(method_name,)` or `(method_name, original_func)`.
            if len(attr) == 1:
                # Delete attribute that did not exist before.
                delattr(cls, attr[0])
            else:
                # Restore original attribute.
                setattr(cls, attr[0], attr[1])

    return _unapply_overrides



# Copyright 2023 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import asyncio
import json
from abc import abstractmethod

import requests

from .._dispatcher_plugins import local
from .._shared_files import logger
from .._shared_files.config import get_config
from .._shared_files.util_classes import RESULT_STATUS, Status

app_log = logger.app_log
log_stack_info = logger.log_stack_info


class BaseTrigger:
    """
    Base class to be subclassed by any custom defined trigger.
    Implements all the necessary methods used for interacting with dispatches, including
    getting their statuses and performing a redispatch of them whenever the trigger gets triggered.

    Args:
        lattice_dispatch_id: Dispatch ID of the worfklow which has to be redispatched in case this trigger gets triggered
        dispatcher_addr: Address of dispatcher server used to retrieve info about or redispatch any dispatches
        triggers_server_addr: Address of the Triggers server (if there is any) to register this trigger to,
                              uses the dispatcher's address by default

    Attributes:
        self.lattice_dispatch_id: Dispatch ID of the worfklow which has to be redispatched in case this trigger gets triggered
        self.dispatcher_addr: Address of dispatcher server used to retrieve info about or redispatch any dispatches
        self.triggers_server_addr: Address of the Triggers server (if there is any) to register this trigger to,
                              uses the dispatcher's address by default
        self.new_dispatch_ids: List of all the newly created dispatch ids from performing redispatch
        self.observe_blocks: Boolean to indicate whether the `self.observe` method is a blocking call
        self.event_loop: Event loop to be used if directly calling dispatcher's functions instead of the REST APIs
        self.use_internal_funcs: Boolean indicating whether to use dispatcher's functions directly instead of through API calls
        self.stop_flag: To handle stopping mechanism in a thread safe manner in case `self.observe()` is a blocking call (e.g. see TimeTrigger)
    """

    def __init__(
        self,
        lattice_dispatch_id: str = None,
        dispatcher_addr: str = None,
        triggers_server_addr: str = None,
    ):
        self.lattice_dispatch_id = lattice_dispatch_id
        self.dispatcher_addr = dispatcher_addr
        self.triggers_server_addr = triggers_server_addr
        self.new_dispatch_ids = []
        self.observe_blocks = True
        self.event_loop = (
            None  # to attach the event loop when directly using dispatcher's functions
        )
        self.use_internal_funcs = (
            True  # whether to use dispatcher's functions directly instead of through API calls
        )
        self.stop_flag = None  # to handle stopping mechanism in a thread safe manner in case observe() is a blocking call (e.g. see TimeTrigger)

    def register(self) -> None:
        """
        Register this trigger to the Triggers server and start observing.
        """
        self._register(self.to_dict())

    @staticmethod
    def _register(trigger_data) -> None:
        """
        Register a trigger to the Triggers server given only its dictionary format and start observing.

        Args:
            trigger_data: Dictionary representation of a trigger
        """

        triggers_server_addr = trigger_data.get("triggers_server_addr")
        if triggers_server_addr is None:
            triggers_server_addr = (
                get_config("dispatcher.address") + ":" + str(get_config("dispatcher.port"))
            )
        register_trigger_url = f"http://{triggers_server_addr}/api/triggers/register"

        r = requests.post(register_trigger_url, json=trigger_data)
        r.raise_for_status()

    def _get_status(self) -> Status:
        """
        Get status about the connected dispatch id to check whether its a pending
        dispatch or new redispatch has to be made.

        Returns:
            status: Status
        """

        if self.use_internal_funcs:
            from covalent_dispatcher._service.app import export_result

            response = asyncio.run_coroutine_threadsafe(
                export_result(self.lattice_dispatch_id, status_only=True),
                self.event_loop,
            ).result()

            if isinstance(response, dict):
                return response["status"]

            return json.loads(response.body.decode()).get("status")

        from .. import get_result

        return get_result(
            self.lattice_dispatch_id, status_only=True, dispatcher_addr=self.dispatcher_addr
        )["status"]

    def _do_redispatch(self, is_pending: bool = False) -> str:
        """
        Perform a redispatch of the connected dispatch id and return a new one.

        Args:
            is_pending: Whether the connected dispatch id is pending

        Returns:
            new_dispatch_id: Dispatch id of the newly dispatched workflow
        """

        if is_pending:
            return local.LocalDispatcher.start(self.lattice_dispatch_id, self.dispatcher_addr)
        else:
            return local.LocalDispatcher.redispatch(
                self.lattice_dispatch_id, self.dispatcher_addr
            )()

    def trigger(self) -> None:
        """
        Trigger this trigger and perform a redispatch of the connected dispatch id's workflow.
        Should be called within `self.observe()` whenever a trigger action is desired.

        Raises:
            RuntimeError: In case no dispatch id is connected to this trigger
        """

        if not self.lattice_dispatch_id:
            raise RuntimeError(
                "`lattice_dispatch_id` is None. Please attach this trigger to a lattice before triggering."
            )

        status = self._get_status()

        if status == str(RESULT_STATUS.NEW_OBJECT) or status is None:
            # To continue the pending dispatch
            same_dispatch_id = self._do_redispatch(True)
            app_log.debug(f"Initiating run for pending dispatch_id: {same_dispatch_id}")
        else:
            # To run new redispatch
            new_dispatch_id = self._do_redispatch()
            app_log.debug(f"Redispatching, new dispatch_id: {new_dispatch_id}")
            self.new_dispatch_ids.append(new_dispatch_id)

    def to_dict(self) -> dict:
        """
        Return a dictionary representation of this trigger which can later be used to regenerate it.

        Returns:
            tr_dict: Dictionary representation of this trigger
        """

        tr_dict = self.__dict__.copy()
        tr_dict["name"] = str(self.__class__.__name__)
        return tr_dict

    @abstractmethod
    def observe(self):
        """
        Start observing for any change which can be used to trigger this trigger.
        To be implemented by the subclass.
        """
        pass

    @abstractmethod
    def stop(self):
        """
        Stop observing for changes.
        To be implemented by the subclass.
        """
        pass



# Copyright 2023 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Defines the core functionality of the new improved runner
"""

import asyncio
import traceback
from concurrent.futures import ThreadPoolExecutor
from datetime import datetime, timezone
from typing import Any, Dict

from covalent._shared_files import logger
from covalent._shared_files.config import get_config
from covalent._shared_files.exceptions import TaskCancelledError
from covalent._shared_files.util_classes import RESULT_STATUS
from covalent.executor.base import AsyncBaseExecutor
from covalent.executor.schemas import ResourceMap, TaskSpec
from covalent.executor.utils import Signals

from . import data_manager as datamgr
from . import runner as runner_legacy
from .data_modules import asset_manager as am
from .runner_modules import executor_proxy, jobs
from .runner_modules.cancel import cancel_tasks  # nopycln: import
from .runner_modules.utils import get_executor

app_log = logger.app_log
log_stack_info = logger.log_stack_info
debug_mode = get_config("sdk.log_level") == "debug"

# Dedicated thread pool for invoking non-async Executor.cancel()
_cancel_threadpool = ThreadPoolExecutor()

# Asyncio Queue
_job_events = None

_job_event_listener = None

_futures = set()

# message format:
# Ready for retrieve result
# {"task_group_metadata": dict, "event": "READY"}
#
# Unable to retrieve result (e.g. credentials expired)
#
# {"task_group_metadata": dict, "event": "FAILED", "detail": str}


# Domain: runner
async def _submit_abstract_task_group(
    dispatch_id: str,
    task_group_id: int,
    task_seq: list,
    known_nodes: list,
    executor: AsyncBaseExecutor,
) -> None:
    # Task sequence of the form {"function_id": task_id, "args_ids":
    # [node_ids], "kwargs_ids": {key: node_id}}

    task_ids = [task["function_id"] for task in task_seq]
    task_specs = []
    task_group_metadata = {
        "dispatch_id": dispatch_id,
        "task_group_id": task_group_id,
        "node_ids": task_ids,
    }

    try:
        if not type(executor).SUPPORTS_MANAGED_EXECUTION:
            raise NotImplementedError("Executor does not support managed execution")

        resources = {"functions": {}, "inputs": {}, "hooks": {}}

        # Get upload URIs
        for task_spec in task_seq:
            task_id = task_spec["function_id"]

            function_uri = executor.get_upload_uri(task_group_metadata, f"function-{task_id}")
            hooks_uri = executor.get_upload_uri(task_group_metadata, f"hooks-{task_id}")

            await am.upload_asset_for_nodes(dispatch_id, "function", {task_id: function_uri})
            await am.upload_asset_for_nodes(dispatch_id, "hooks", {task_id: hooks_uri})

            resources["functions"][task_id] = function_uri
            resources["hooks"][task_id] = hooks_uri

            task_specs.append(TaskSpec(**task_spec))

        node_upload_uris = {
            node_id: executor.get_upload_uri(task_group_metadata, f"node_{node_id}")
            for node_id in known_nodes
        }
        resources["inputs"] = node_upload_uris

        app_log.debug(
            f"Uploading known nodes {known_nodes} for task group {dispatch_id}:{task_group_id}"
        )
        await am.upload_asset_for_nodes(dispatch_id, "output", node_upload_uris)

        ts = datetime.now(timezone.utc)
        node_results = [
            datamgr.generate_node_result(
                node_id=task_id,
                start_time=ts,
                status=RESULT_STATUS.RUNNING,
            )
            for task_id in task_ids
        ]

        # Use one proxy for the task group; handles the following requests:
        # - check if the job has a pending cancel request
        # - set the job handle
        # - set job status

        # Watch the task group
        fut = asyncio.create_task(executor_proxy.watch(dispatch_id, task_ids[0], executor))
        _futures.add(fut)
        fut.add_done_callback(_futures.discard)

        send_retval = await executor.send(
            task_specs,
            ResourceMap(**resources),
            task_group_metadata,
        )

        app_log.debug(f"Submitted task group {dispatch_id}:{task_group_id}")

    except TaskCancelledError:
        app_log.debug(f"Task group {dispatch_id}:{task_group_id} cancelled")

        send_retval = None

        node_results = [
            datamgr.generate_node_result(
                node_id=task_id,
                end_time=datetime.now(timezone.utc),
                status=RESULT_STATUS.CANCELLED,
            )
            for task_id in task_ids
        ]

    except Exception as ex:
        tb = "".join(traceback.TracebackException.from_exception(ex).format())
        app_log.debug(f"Exception occurred when running task group {task_group_id}:")
        app_log.debug(tb)
        error_msg = tb if debug_mode else str(ex)
        ts = datetime.now(timezone.utc)

        send_retval = None

        node_results = [
            datamgr.generate_node_result(
                node_id=task_id,
                end_time=datetime.now(timezone.utc),
                status=RESULT_STATUS.FAILED,
                error=error_msg,
            )
            for task_id in task_ids
        ]

    return node_results, send_retval


async def _get_task_result(task_group_metadata: Dict, data: Any):
    """Retrieve task results from executor.

    Parameters:
        task_group_metadata: metadata about the task group
        data: task execution information (such as status)

    Both `task_group_metadata` and `data` will be passed directly to
    Executor.receive().

    """

    dispatch_id = task_group_metadata["dispatch_id"]
    task_ids = task_group_metadata["node_ids"]
    gid = task_group_metadata["task_group_id"]
    app_log.debug(f"Pulling job artifacts for task group {dispatch_id}:{gid}")
    try:
        executor_attrs = await datamgr.electron.get(
            dispatch_id, gid, ["executor", "executor_data"]
        )
        executor_name = executor_attrs["executor"]
        executor_data = executor_attrs["executor_data"]

        executor = get_executor(
            node_id=gid,
            selected_executor=[executor_name, executor_data],
            loop=asyncio.get_running_loop(),
            pool=None,
        )

        # Expects a list of TaskUpdates
        task_group_results = await executor.receive(task_group_metadata, data)

        node_results = []
        for task_result in task_group_results:
            task_id = task_result.node_id
            status = task_result.status
            await am.download_assets_for_node(dispatch_id, task_id, task_result.assets)

            node_result = datamgr.generate_node_result(
                node_id=task_id,
                end_time=datetime.now(timezone.utc),
                status=status,
            )
            node_results.append(node_result)

    except Exception as ex:
        tb = "".join(traceback.TracebackException.from_exception(ex).format())
        app_log.debug(f"Exception occurred when receiving task group {gid}:")
        app_log.debug(tb)
        error_msg = tb if debug_mode else str(ex)
        ts = datetime.now(timezone.utc)

        node_results = [
            datamgr.generate_node_result(
                node_id=node_id,
                end_time=ts,
                status=RESULT_STATUS.FAILED,
                error=error_msg,
            )
            for node_id in task_ids
        ]

    for node_result in node_results:
        await datamgr.update_node_result(dispatch_id, node_result)


async def run_abstract_task_group(
    dispatch_id: str,
    task_group_id: int,
    task_seq: list,
    known_nodes: list,
    selected_executor: Any,
) -> None:
    executor = None

    try:
        app_log.debug(f"Attempting to instantiate executor {selected_executor}")
        task_ids = [task["function_id"] for task in task_seq]
        app_log.debug(f"Running task group {dispatch_id}:{task_group_id}")
        executor = get_executor(
            node_id=task_group_id,
            selected_executor=selected_executor,
            loop=asyncio.get_running_loop(),
            pool=None,
        )

        # Check if the job should be cancelled
        if await jobs.get_cancel_requested(dispatch_id, task_ids[0]):
            await jobs.put_job_status(dispatch_id, task_ids[0], RESULT_STATUS.CANCELLED)

            for task_id in task_ids:
                task_metadata = {"dispatch_id": dispatch_id, "node_id": task_id}
                app_log.debug(f"Refusing to execute cancelled task {dispatch_id}:{task_id}")
                await mark_task_ready(task_metadata, None)

            return

        # Legacy runner doesn't yet support task packing
        if not type(executor).SUPPORTS_MANAGED_EXECUTION:
            if len(task_seq) != 1:
                raise RuntimeError("Task packing not supported by executor plugin")

            task_spec = task_seq[0]
            node_id = task_spec["function_id"]
            name = task_spec["name"]
            abstract_inputs = {
                "args": task_spec["args_ids"],
                "kwargs": task_spec["kwargs_ids"],
            }
            app_log.debug(f"Reverting to legacy runner for task {task_group_id}")
            coro = runner_legacy.run_abstract_task(
                dispatch_id,
                node_id,
                name,
                abstract_inputs,
                selected_executor,
            )
            fut = asyncio.create_task(coro)
            _futures.add(fut)
            fut.add_done_callback(_futures.discard)
            return

        node_results, send_retval = await _submit_abstract_task_group(
            dispatch_id,
            task_group_id,
            task_seq,
            known_nodes,
            executor,
        )

    except Exception as ex:
        tb = "".join(traceback.TracebackException.from_exception(ex).format())
        app_log.debug("Exception when trying to instantiate executor:")
        app_log.debug(tb)
        error_msg = tb if debug_mode else str(ex)
        ts = datetime.now(timezone.utc)

        send_retval = None
        node_results = [
            datamgr.generate_node_result(
                node_id=node_id,
                start_time=ts,
                end_time=ts,
                status=RESULT_STATUS.FAILED,
                error=error_msg,
            )
            for node_id in task_ids
        ]

    for node_result in node_results:
        await datamgr.update_node_result(dispatch_id, node_result)

    if node_results[0]["status"] == RESULT_STATUS.RUNNING:
        task_group_metadata = {
            "dispatch_id": dispatch_id,
            "node_ids": task_ids,
            "task_group_id": task_group_id,
        }
        await _poll_task_status(task_group_metadata, executor, send_retval)

    # Terminate proxy
    if executor:
        executor._notify(Signals.EXIT)
        app_log.debug(f"Stopping proxy for task group {dispatch_id}:{task_group_id}")


async def _listen_for_job_events():
    app_log.debug("Starting event listener")
    while True:
        msg = await _job_events.get()
        try:
            event = msg["event"]
            app_log.debug(f"Received job event {event}")
            if event == "BYE":
                app_log.debug("Terminating job event listener")
                break

            # job has reached a terminal state
            if event == "READY":
                task_group_metadata = msg["task_group_metadata"]
                detail = msg["detail"]
                fut = asyncio.create_task(_get_task_result(task_group_metadata, detail))
                _futures.add(fut)
                fut.add_done_callback(_futures.discard)
                continue

            if event == "FAILED":
                task_group_metadata = msg["task_group_metadata"]
                dispatch_id = task_group_metadata["dispatch_id"]
                gid = task_group_metadata["task_group_id"]
                task_ids = task_group_metadata["node_ids"]
                detail = msg["detail"]
                ts = datetime.now(timezone.utc)
                for task_id in task_ids:
                    node_result = datamgr.generate_node_result(
                        node_id=task_id,
                        end_time=ts,
                        status=RESULT_STATUS.FAILED,
                        error=detail,
                    )
                    await datamgr.update_node_result(dispatch_id, node_result)

        except Exception as ex:
            app_log.exception("Error reading message: {ex}")


async def _mark_ready(task_group_metadata: dict, detail: Any):
    await _job_events.put(
        {"task_group_metadata": task_group_metadata, "event": "READY", "detail": detail}
    )


async def _mark_failed(task_group_metadata: dict, detail: str):
    await _job_events.put(
        {"task_group_metadata": task_group_metadata, "event": "FAILED", "detail": detail}
    )


async def _poll_task_status(
    task_group_metadata: Dict, executor: AsyncBaseExecutor, poll_data: Any
):
    """Polls a group of tasks until it terminates.

    `poll_data` is the return value of `Executor.send()` and will be
    passed directly to `Executor.poll()`.

    """
    # Return immediately if no polling logic (default return value is -1)

    dispatch_id = task_group_metadata["dispatch_id"]
    task_group_id = task_group_metadata["task_group_id"]
    task_ids = task_group_metadata["node_ids"]

    try:
        app_log.debug(f"Polling status for task group {dispatch_id}:{task_group_id}")
        receive_data = await executor.poll(task_group_metadata, poll_data)
        await _mark_ready(task_group_metadata, receive_data)

    except NotImplementedError:
        app_log.debug(f"Executor {executor.short_name()} is async.")

    except TaskCancelledError:
        app_log.debug(f"Task group {dispatch_id}:{task_group_id} cancelled")
        await _mark_ready(task_group_metadata, None)

    except Exception as ex:
        task_group_id = task_group_metadata["task_group_id"]
        tb = "".join(traceback.TracebackException.from_exception(ex).format())
        app_log.debug(f"Exception occurred when polling task {task_group_id}:")
        app_log.debug(tb)
        error_msg = tb if debug_mode else str(ex)
        await _mark_failed(task_group_metadata, error_msg)


async def mark_task_ready(task_metadata: dict, detail: Any):
    dispatch_id = task_metadata["dispatch_id"]
    node_id = task_metadata["node_id"]
    gid = (await datamgr.electron.get(dispatch_id, node_id, ["task_group_id"]))["task_group_id"]
    task_group_metadata = {
        "dispatch_id": dispatch_id,
        "node_ids": [node_id],
        "task_group_id": gid,
    }

    await _mark_ready(task_group_metadata, detail)



# Copyright 2023 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""""
Pennylane-Qiskit device that uses the Qiskit Runtime `Sampler` primitive
"""

from typing import Any, List, Sequence, Union

import pennylane as qml
from devices_base import QiskitSamplerDevice
from qiskit_ibm_runtime import Sampler
from qiskit_utils import extract_options
from sessions import get_cached_session


class QiskitRuntimeSampler(QiskitSamplerDevice):
    """
    Pennylane device that runs circuits with Qiskit Runtime's `Sampler`
    """

    short_name = "sampler"

    def __init__(  # pylint: disable=too-many-arguments
        self,
        wires: int,
        shots: Union[None, int, Sequence[int], Sequence[Union[int, Sequence[int]]]],
        backend_name: str,
        local_transpile: bool,
        max_time: Union[int, str],
        single_job: bool,
        options: dict,
        service_init_kwargs: dict,
    ):
        if options:
            _options = extract_options(options)
            _options.execution.shots = shots
        else:
            _options = None

        self.options = _options
        self.max_time = max_time
        self.single_job = single_job

        # These attributes are shared across `batch_execute` and `post_process`
        self._active_jobs = []
        self._active_circuits = []
        self._vector_input = False

        super().__init__(
            wires=wires,
            shots=shots,
            backend_name=backend_name,
            local_transpile=local_transpile,
            service_init_kwargs=service_init_kwargs,
        )

    def batch_execute(self, circuits, timeout: int = None):
        """
        Submit a batch of circuits to IBM Qiskit Runtime for execution.
        """

        self._n_original_circuits = len(circuits)
        circuits = self.broadcast_tapes(circuits)
        self._n_circuits = len(circuits)

        # Create circuit objects and apply diagonalizing gates
        compiled_circuits = self.compile_circuits(circuits)
        self._active_circuits = [circuit.copy() for circuit in circuits]

        # Submit circuit objects to Qiskit Runtime
        max_time = timeout or self.max_time
        with get_cached_session(self.service, self.backend, max_time) as session:
            sampler = Sampler(session=session, options=self.options)

            if self.single_job:
                job = sampler.run(compiled_circuits)
                self._active_jobs.append(job)
            else:
                for compiled_circuit in compiled_circuits:
                    job = sampler.run(compiled_circuit)
                    self._active_jobs.append(job)

        # This flag distinguishes vector inputs from gradient computations
        self._vector_input = self._n_original_circuits != self._n_circuits

        # Return a dummy result and proceed to subsequent submissions.
        return self._dummy_result()

    def post_process(self, *args):
        """
        Post-process a single circuit result.
        """

        results = []
        metadatas = []
        for i, circuit in enumerate(self._active_circuits):
            # Get broadcasted circuit and active job
            circuit = self._active_circuits[i]
            job = self._active_jobs[i]

            # Blocking call to retrieve job result
            job_result = job.result()
            self._num_executions += 1

            # Extra metadata and quasi-distribution
            assert len(job_result.quasi_dists) == 1
            assert len(job_result.metadata) == 1
            quasi_dist = job_result.quasi_dists[0]

            # Post process the quasi-distribution into expected result
            res = self._process_batch_execute_result(circuit, quasi_dist)
            results.append(res)

            # Update metadata
            job_metadata = job_result.metadata[0]
            metadata = {
                "result_object": job_result,
                "num_measurements": self.num_executions,
            }
            metadata.update(**job_metadata)
            metadatas.append(metadata)

        # Update tracker
        if self.tracker.active:
            self.tracker.update(batches=len(self._active_circuits), batch_len=1)
            self.tracker.record()

        # Reset active jobs and circuits
        self._active_jobs = []
        self._active_circuits = []

        # Re-execute original tape with a dummy device to get correct result
        tape = args[0]
        results = self._re_execute(self.broadcast_tapes([tape]), results)

        # Wrap in outer list for vector inputs
        if self._vector_input:
            results = self._vector_results(results)

        return results, metadatas

    def post_process_all(self, *args):
        """
        Post-process a batch of circuit results.
        """

        # Blocking call to retrieve job result
        job = self._active_jobs.pop()
        job_result = job.result()

        # Increment counter for number of executions of qubit device
        self._num_executions += 1

        # Compute statistics using the state and/or samples
        results = []
        metadatas = []

        for i, circuit in enumerate(self._active_circuits):
            quasi_dist = job_result.quasi_dists[i]

            # Update tracker and process quasi-distribution into expected result
            res = self._process_batch_execute_result(circuit, quasi_dist)
            results.append(res)

            # Construct metadata
            job_metadata = job_result.metadata[i]
            metadata = {
                "result_object": job_result,
                "num_measurements": self._num_executions,
            }
            metadata.update(**job_metadata)
            metadatas.append(metadata)

        # Update tracker
        if self.tracker.active:
            self.tracker.update(batches=1, batch_len=len(self._active_circuits))
            self.tracker.record()

        # Reset active jobs and circuits
        self._active_jobs = []
        self._active_circuits = []

        # Re-execute with a dummy device to get correct result
        tapes = args[0]
        results = self._re_execute(self.broadcast_tapes(tapes), results)

        # Wrap in outer list for vector inputs
        if self._vector_input:
            results = self._vector_results(results)

        return results, metadatas

    def _re_execute(self, tapes, results):
        """
        Executes circuits on a dummy device that returns the provided result.

        This is necessary for the raw output from `post_process` to be handled
        as if it came from `batch_execute`.
        """
        dev = _PostProcessDevice(self.wires, results)
        return qml.execute(tapes, dev, gradient_fn="best")

    def _dummy_result(self) -> Union[Any, List[Any]]:
        """
        Returns a dummy result to satisfy the expected `qml.execute` output.
        This allows async submission to proceed without waiting for the job result.
        """

        dummy_results = qml.numpy.zeros(self.vector_shape)

        if self.pennylane_active_return:
            dummy_results = list(dummy_results)

        if self.single_job and self._vector_input:
            if self.pennylane_active_return:
                return [[d] for d in dummy_results]
            return [[d] for d in list(dummy_results)]

        if self._vector_input:
            return [dummy_results]
        return [[d] for d in dummy_results]


class _PostProcessDevice(QiskitRuntimeSampler):
    """
    A copy of the QiskitRuntimeSampler class with a dummy `batch_execute` method
    that returns the assigned `results`.
    """

    def __init__(self, wires, results: Any):
        super().__init__(
            wires=wires,
            shots=1,
            backend_name="",
            local_transpile=False,
            max_time=1,
            single_job=False,
            options={},
            service_init_kwargs={},
        )

        self._results = results

    def batch_execute(self, *_, **__):
        """
        Override to return expected result.
        """
        return self._results



# Copyright 2023 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
from typing import Callable, List, Optional, Union

import pennylane as qml

from .._shared_files.qelectron_utils import get_import_path, get_original_shots
from ..quantum.qcluster import QCluster
from ..quantum.qcluster.base import AsyncBaseQCluster, BaseQExecutor
from ..quantum.qcluster.simulator import Simulator
from .qnode import QElectronInfo, QNodeQE

Selector = Union[str, Callable[[qml.tape.QuantumScript, List[BaseQExecutor]], BaseQExecutor]]


def qelectron(
    qnode: Optional[qml.QNode] = None,
    *,
    executors: Union[BaseQExecutor, AsyncBaseQCluster, List[BaseQExecutor]] = None,
    name: Optional[str] = None,
    description: Optional[str] = None,
    selector: Union[str, Selector] = "cyclic",
) -> QNodeQE:
    """
    QElectron decorator to be called upon a Pennylane QNode. Adds multi-backend
    execution functionality to the original QNode.

    Args:
        qnode: The Pennylane :code:`QNode` to wrap.

    Keyword Args:
        executors: The quantum executor(s) to use for running the QNode. A single
            executor, list of executors, or a :code:`QCluster` instance are accepted.
            If a list of multiple executors is passed, a quantum cluster is
            initialized from this list automatically and :code:`selector` is used as the
            cluster's selector. Defaults to a thread-based :code:`Simulator`.
        name: An optional name for the QElectron. Defaults to the circuit function's
            name.
        description: An optional description for the QElectron. Defaults to the
            circuit function's docstring.
        selector: A callable that selects an executor, or one of the strings :code:`"cyclic"`
            or :code:`"random"`. The :code:`"cyclic"` selector (default) cycles through
            :code:`executors` and returns the next executor for each circuit. The
            :code:`"random"` selector chooses an executor from :code:`executors`
            at random for each circuit. Any user-defined selector must be callable
            with two positional arguments, a circuit and a list of executors.
            A selector must also return exactly one executor.

    Raises:
        ValueError: If any invalid executors are passed.

    Returns:
        :code:`QNodeQE`: A sub-type of :code:`QNode` that integrates QElectrons.
    """

    if executors is None:
        executors = Simulator()

    # Check if executor is a list of executors.
    if isinstance(executors, list):
        if not all(isinstance(ex, BaseQExecutor) for ex in executors):
            raise ValueError("Invalid executor in executors list.")
        if len(executors) > 1:
            # Convert to cluster if more than one executor in list.
            executors = QCluster(executors=executors, selector=selector)

    # Check if executor is a QCluster.
    if isinstance(executors, AsyncBaseQCluster):
        # Serialize the cluster's selector function.
        executors.serialize_selector()

    # Check if a single executor instance was passed.
    if not isinstance(executors, list):
        executors = [executors]

    if qnode is None:
        # This only happens when `qelectron()` is not used as a decorator.
        return functools.partial(
            qelectron, executors=executors, name=name, description=description
        )

    # Set default name and description.
    if name is None:
        name = qnode.func.__name__

    if description is None:
        description = qnode.func.__doc__

    qelectron_info = get_qinfo(name, description, qnode)

    # Create and return a new `QNodeQE` instance.
    return QNodeQE(qnode, executors, qelectron_info)


def get_qinfo(name: str, description: str, qnode: qml.QNode) -> QElectronInfo:
    """
    Extract useful information about the original QNode.
    """

    device_import_path = get_import_path(type(qnode.device))
    device_shots = get_original_shots(qnode.device)
    device_shots_type = None if device_shots is None else type(device_shots)

    return QElectronInfo(
        name=name,
        description=description,
        device_name=qnode.device.short_name,
        device_import_path=device_import_path,
        device_shots=device_shots,
        device_shots_type=device_shots_type,
        device_wires=qnode.device.num_wires,
        pennylane_active_return=qml.active_return(),
    )



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Self-contained entry point for the dispatcher
"""

import asyncio
from typing import List, Optional

from covalent._shared_files import logger
from covalent._shared_files.schemas.result import ResultSchema

from ._core import cancel_dispatch

app_log = logger.app_log
log_stack_info = logger.log_stack_info


async def make_dispatch(json_lattice: str):
    """
    Run the dispatcher from the lattice asynchronously using Dask.
    Assign a new dispatch id to the result object and return it.
    Also save the result in this initial stage to the file mentioned in the result object.

    Args:
        json_lattice: A JSON-serialized lattice

    Returns:
        dispatch_id: A string containing the dispatch id of current dispatch.
    """

    from ._core import make_dispatch

    dispatch_id = await make_dispatch(json_lattice)

    app_log.debug(f"Created new dispatch {dispatch_id}")

    return dispatch_id


async def start_dispatch(dispatch_id: str):
    """
    Run the dispatcher from the lattice asynchronously using Dask.
    Assign a new dispatch id to the result object and return it.
    Also save the result in this initial stage to the file mentioned in the result object.

    Args:
        json_lattice: A JSON-serialized lattice

    Returns:
        dispatch_id: A string containing the dispatch id of current dispatch.
    """

    from ._core import copy_futures, run_dispatch

    # Wait for any pending asset transfers
    _fut = copy_futures.get(dispatch_id, None)
    if _fut is not None:
        # _fut is a concurrent.future.Future, so we need to wrap it in
        # an asyncio.Future
        app_log.debug(f"Waiting on asset transfers for dispatch {dispatch_id}")
        await asyncio.wrap_future(_fut)

    # Idempotent
    run_dispatch(dispatch_id)
    app_log.debug(f"Running dispatch {dispatch_id}")


async def run_dispatcher(json_lattice: str):
    """
    Run the dispatcher from the lattice asynchronously using Dask.
    Assign a new dispatch id to the result object and return it.
    Also save the result in this initial stage to the file mentioned in the result object.

    Args:
        json_lattice: A JSON-serialized lattice

    Returns:
        dispatch_id: A string containing the dispatch id of current dispatch.
    """

    dispatch_id = await make_dispatch(json_lattice)
    await start_dispatch(dispatch_id)

    app_log.debug("Submitted result object to run_workflow.")

    return dispatch_id


async def cancel_running_dispatch(dispatch_id: str, task_ids: List[int] = None) -> None:
    """
    Cancels a running dispatch job.

    Args:
        dispatch_id: Dispatch id of the dispatch to be cancelled.

    Returns:
        None
    """

    if task_ids is None:
        task_ids = []

    await cancel_dispatch(dispatch_id, task_ids)


async def register_dispatch(
    manifest: ResultSchema, parent_dispatch_id: Optional[str]
) -> ResultSchema:
    from ._core.data_modules.importer import import_manifest

    return await import_manifest(manifest, parent_dispatch_id, None)


async def register_redispatch(
    manifest: ResultSchema,
    parent_dispatch_id: str,
    reuse_previous_results: bool,
) -> ResultSchema:
    from ._core.data_modules.importer import import_derived_manifest

    return await import_derived_manifest(
        manifest,
        parent_dispatch_id,
        reuse_previous_results,
    )



# Copyright 2023 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""FastAPI models for /api/v1/resultv2 endpoints"""

from enum import Enum

from covalent._shared_files.util_classes import RESULT_STATUS


class StatusEnum(str, Enum):
    NEW_OBJECT = str(RESULT_STATUS.NEW_OBJECT)
    STARTING = str(RESULT_STATUS.STARTING)
    PENDING_REUSE = str(RESULT_STATUS.PENDING_REUSE)  # For redispatch in the new dispatcher design
    PENDING_REPLACEMENT = str(
        RESULT_STATUS.PENDING_REPLACEMENT
    )  # For redispatch in the new dispatcher design
    COMPLETED = str(RESULT_STATUS.COMPLETED)
    POSTPROCESSING = str(RESULT_STATUS.POSTPROCESSING)
    FAILED = str(RESULT_STATUS.FAILED)
    RUNNING = str(RESULT_STATUS.RUNNING)
    CANCELLED = str(RESULT_STATUS.CANCELLED)
    DISPATCHING = str(RESULT_STATUS.DISPATCHING)



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
from datetime import datetime, timezone

import aiofiles

from covalent._shared_files.config import get_config


class Heartbeat:
    TIMESTAMP_FORMAT = "%Y-%m-%d %H:%M:%S.%f%z"

    def __init__(
        self,
        beat_interval: int = None,
        beat_file: str = None,
    ):
        self.beat_interval = beat_interval or get_config("dispatcher.heartbeat_interval")
        self.beat_file = beat_file or get_config("dispatcher.heartbeat_file")

    async def start(self):
        while True:
            await self.beat()
            await asyncio.sleep(self.beat_interval)

    async def beat(self):
        async with aiofiles.open(self.beat_file, mode="w") as file:
            await file.write(
                f"ALIVE {datetime.now(tz=timezone.utc).strftime(Heartbeat.TIMESTAMP_FORMAT)}"
            )

    @staticmethod
    def stop():
        with open(get_config("dispatcher.heartbeat_file"), mode="w") as file:
            file.write(
                f"DEAD {datetime.now(tz=timezone.utc).strftime(Heartbeat.TIMESTAMP_FORMAT)}"
            )



# Copyright 2023 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Utilities for querying the transport graph
"""


# Note: these query static information which should be amenable to caching

from typing import Dict, List

import networkx as nx

from ..._dal.result import get_result_object
from .utils import run_in_executor


def get_incoming_edges_sync(dispatch_id: str, node_id: int):
    """Query in-edges of a node.

    Returns:
        List[Edge], where

        Edge is a dictionary with structure
            source: int,
            target: int,
            attrs: dict
    """

    result_object = get_result_object(dispatch_id)
    return result_object.lattice.transport_graph.get_incoming_edges(node_id)


def get_node_successors_sync(
    dispatch_id: str,
    node_id: int,
    attrs: List[str],
) -> List[Dict]:
    """Get child nodes with multiplicity.

    Parameters:
        node_id: id of node
        attr_keys: list of node attributes to return, such as task_group_id

    Returns:
        List[Dict], where each dictionary is of the form
        {"node_id": node_id, attr_key_1: node_attr[attr_key_1], ...}

    """

    result_object = get_result_object(dispatch_id)
    return result_object.lattice.transport_graph.get_successors(node_id, attrs)


def get_nodes_links_sync(dispatch_id: str) -> dict:
    """Return the internal transport graph in NX node-link form"""

    # Need the whole NX graph here
    result_object = get_result_object(dispatch_id, False)
    g = result_object.lattice.transport_graph.get_internal_graph_copy()
    return nx.readwrite.node_link_data(g)


def get_nodes_sync(dispatch_id: str) -> List[int]:
    """Return a list of all node ids in the graph."""
    result_object = get_result_object(dispatch_id, False)
    g = result_object.lattice.transport_graph.get_internal_graph_copy()
    return list(g.nodes)


async def get_incoming_edges(dispatch_id: str, node_id: int):
    return await run_in_executor(get_incoming_edges_sync, dispatch_id, node_id)


async def get_node_successors(
    dispatch_id: str,
    node_id: int,
    attrs: List[str] = ["task_group_id"],
) -> List[Dict]:
    return await run_in_executor(get_node_successors_sync, dispatch_id, node_id, attrs)


async def get_nodes_links(dispatch_id: str) -> Dict:
    return await run_in_executor(get_nodes_links_sync, dispatch_id)


async def get_nodes(dispatch_id: str) -> List[int]:
    return await run_in_executor(get_nodes_sync, dispatch_id)



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Graph Route"""

import uuid

from fastapi import APIRouter, HTTPException, status
from fastapi.encoders import jsonable_encoder
from sqlalchemy.orm import Session

import covalent_ui.api.v1.database.config.db as db
from covalent_ui.api.v1.data_layer.graph_dal import Graph
from covalent_ui.api.v1.models.graph_model import GraphResponse

routes: APIRouter = APIRouter()


@routes.get("/{dispatch_id}/graph", response_model=GraphResponse)
def get_graph(dispatch_id: uuid.UUID):
    """Get Graph

    Args:
        dispatch_id: To fetch lattice data with the provided dispatch id

    Returns:
        Returns the lattice data with the dispatch id provided
    """

    with Session(db.engine) as session:
        graph = Graph(session)
        graph_data = graph.get_graph(dispatch_id)
        if graph_data is not None:
            return GraphResponse(
                dispatch_id=graph_data["dispatch_id"],
                graph={
                    "nodes": jsonable_encoder(graph_data["nodes"]),
                    "links": jsonable_encoder(graph_data["links"]),
                },
            )
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail=[
                {
                    "loc": ["path", "dispatch_id"],
                    "msg": f"Dispatch ID {dispatch_id} does not exist",
                    "type": None,
                }
            ],
        )



# Copyright 2023 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time
from functools import partial
from threading import Event
from typing import List

from sqlalchemy import create_engine
from sqlalchemy.orm import Session

from covalent._shared_files import logger

from .base import BaseTrigger

app_log = logger.app_log
log_stack_info = logger.log_stack_info


class DatabaseTrigger(BaseTrigger):
    """
    Database trigger which can read for changes in a database
    and trigger workflows based on record changes.

    Args:
        db_path: Connection string for the database
        table_name: Name of the table to observe
        poll_interval: Time in seconds to wait for before reading the database again
        where_clauses: List of "WHERE" conditions, e.g. ["id > 2", "status = pending"], to check when
                       polling the database
        trigger_after_n: Number of times the event must happen after which the workflow will be triggered.
                         e.g value of 2 means workflow will be triggered once the event has occurred twice.
        lattice_dispatch_id: Lattice dispatch id of the workflow to be triggered
        dispatcher_addr: Address of the dispatcher server
        triggers_server_addr: Address of the triggers server

    Attributes:
        self.db_path: Connection string for the database
        self.table_name: Name of the table to observe
        self.poll_interval: Time in seconds to wait for before reading the database again
        self.where_clauses: List of "WHERE" conditions, e.g. ["id > 2", "status = pending"], to check when
                            polling the database
        self.trigger_after_n: Number of times the event must happen after which the workflow will be triggered.
                              e.g value of 2 means workflow will be triggered once the event has occurred twice.
        self.stop_flag: Thread safe flag used to check whether the stop condition has been met

    """

    def __init__(
        self,
        db_path: str,
        table_name: str,
        poll_interval: int = 1,
        where_clauses: List[str] = None,
        trigger_after_n: int = 1,
        lattice_dispatch_id: str = None,
        dispatcher_addr: str = None,
        triggers_server_addr: str = None,
    ):
        super().__init__(lattice_dispatch_id, dispatcher_addr, triggers_server_addr)

        self.db_path = db_path
        self.table_name = table_name
        self.poll_interval = poll_interval

        self.where_clauses = where_clauses

        self.trigger_after_n = trigger_after_n

        self.stop_flag = None

    def observe(self) -> None:
        """
        Keep performing the trigger action as long as
        where conditions are met or until stop has being called
        """

        app_log.debug("Inside DatabaseTrigger's observe")
        event_count = 0

        try:
            self.engine = create_engine(self.db_path)

            with Session(self.engine) as db:
                sql_poll_cmd = f"SELECT * FROM {self.table_name}"

                if self.where_clauses:
                    sql_poll_cmd += " WHERE "
                    sql_poll_cmd += " AND ".join(list(self.where_clauses))
                sql_poll_cmd += ";"

                execute_cmd = partial(db.execute, sql_poll_cmd)
                app_log.debug(f"Poll command: {sql_poll_cmd}")

                self.stop_flag = Event()
                while not self.stop_flag.is_set():
                    # Read the DB with specified command
                    try:
                        app_log.debug("About to execute...")
                        if rows := execute_cmd().all():
                            event_count += 1
                            if event_count == self.trigger_after_n:
                                app_log.debug("Invoking trigger")
                                self.trigger()
                                event_count = 0

                    except Exception:
                        pass

                    time.sleep(self.poll_interval)
        except Exception as e:
            app_log.debug("Failed to observe:")
            raise

    def stop(self) -> None:
        """
        Stop the running `self.observe()` method by setting the `self.stop_flag` flag.
        """

        self.stop_flag.set()



# Copyright 2023 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any

import pennylane as qml
from pennylane.tape import QuantumTape

from .._shared_files.qresult_utils import re_execute
from ..quantum.qclient.core import middleware


class QNodeFutureResult:
    """
    A class that stores the `batch_id` of a batch of circuits submitted to the
    middleware. The `result` method can then be called to retrieve the results.

    Attributes:
        device: The Pennylane device used by the original QNode.
        interface: The interface of the original QNode.
        diff_method: The differentiation method of the original QNode.
        qfunc_output: The return value (measurement definition) of the original QNode.
    """

    def __init__(
        self,
        batch_id: str,
        interface: str,
        original_qnode: qml.QNode,
        original_tape: QuantumTape,
    ):
        """
        Initialize a `QNodeFutureResult` instance.

        Args:
            batch_id: A UUID that identifies a batch of circuits submitted to
                the middleware.
        """
        self.batch_id = batch_id
        self.interface = interface  # NOT the original QNode's interface

        # Required for batch_transforms and correct output typing.
        self.device = original_qnode.device
        self.qnode = original_qnode
        self.tape = original_tape

        self.args = None
        self.kwargs = None
        self._result = None

    def __call__(self, *args, **kwargs):
        """
        Store the arguments and keyword arguments of the original QNode call.
        """
        self.args = args
        self.kwargs = kwargs
        return self

    def result(self) -> Any:
        """
        Retrieve the results for the given `batch_id` from middleware. This method
        is blocking until the results are available.

        Returns:
            The results of the circuit execution.
        """

        if self._result is None:
            # Get raw results from the middleware.
            results = middleware.get_results(self.batch_id)

            # Required correct gradient post-processing in some cases.
            if self.interface == "autograd":
                self._result = results
                res = results[0]

            if self.interface != "numpy":
                interface = self.interface  # re-execute with any non-numpy interface
                res = results[0]  # re-execute with this result

            elif self.qnode.interface is None:
                interface = None
                res = results[0]

            elif self.qnode.interface == "auto":
                interface = "auto"
                res = results

            else:
                # Skip re-execution.
                self._result = results
                return results

            args, kwargs = self.args, self.kwargs
            self._result = re_execute(res, self.qnode, self.tape)(interface, *args, **kwargs)

        return self._result



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utils for migrating legacy (0.110-era) result object to a modern result object."""

import pickle

from covalent._results_manager import Result
from covalent._shared_files import logger
from covalent._shared_files.defaults import (
    attr_prefix,
    electron_dict_prefix,
    electron_list_prefix,
    generator_prefix,
    parameter_prefix,
    subscript_prefix,
)
from covalent._shared_files.utils import get_named_params
from covalent._workflow.electron import to_decoded_electron_collection
from covalent._workflow.lattice import Lattice
from covalent._workflow.transport import TransportableObject, _TransportGraph, encode_metadata

from .._db import update

app_log = logger.app_log
log_stack_info = logger.log_stack_info


def process_node(node: dict) -> dict:
    """Convert a node from a 0.110.2-vintage transport graph

    Args:
        node: dictionary of node attributes

    Returns:
        the converted node attributes
    """

    if "metadata" in node:
        node["metadata"] = encode_metadata(node["metadata"])
        if "deps" not in node["metadata"]:
            node["metadata"]["deps"] = {}
        if "call_before" not in node["metadata"]:
            node["metadata"]["call_before"] = []
        if "call_after" not in node["metadata"]:
            node["metadata"]["call_after"] = []

    node_name = node["name"]

    # encode output, remove "attribute_name", strip "attr_prefix" from name
    if node_name.startswith(attr_prefix):
        node["output"] = TransportableObject.make_transportable(node["output"])
        if "attribute_name" in node:
            del node["attribute_name"]
        new_node_name = node_name.replace(attr_prefix, "")
        node["name"] = new_node_name

    # encode output, remove "key", strip "generator_prefix" from name
    elif node_name.startswith(generator_prefix):
        node["output"] = TransportableObject.make_transportable(node["output"])
        if "key" in node:
            del node["key"]
        new_node_name = node_name.replace(generator_prefix, "")
        node["name"] = new_node_name

    # encode output, remove "key", strip "subscript_prefix" from name
    elif node_name.startswith(subscript_prefix):
        node["output"] = TransportableObject.make_transportable(node["output"])
        if "key" in node:
            del node["key"]
        new_node_name = node_name.replace(subscript_prefix, "")
        node["name"] = new_node_name

    # Replace function for collection nodes
    elif node_name.startswith(electron_list_prefix) or node_name.startswith(electron_dict_prefix):
        node["function"] = TransportableObject(to_decoded_electron_collection)

    # Encode "value" and "output" for parameter nodes
    elif node_name.startswith(parameter_prefix):
        node["value"] = TransportableObject.make_transportable(node["value"])
        node["output"] = TransportableObject.make_transportable(node["output"])

    # Function nodes: encode output and sublattice_result
    else:
        node["output"] = TransportableObject.make_transportable(node["output"])
        if "sublattice_result" in node:
            if node["sublattice_result"] is not None:
                node["sublattice_result"] = process_result_object(node["sublattice_result"])

    return node


def process_transport_graph(tg: _TransportGraph) -> _TransportGraph:
    """Convert a 0.110.2-vintage transport graph to a modern transport graph

    Args:
        tg: old Transport Graph

    Returns:
        the modernized Transport Graph
    """
    tg_new = _TransportGraph()
    g = tg.get_internal_graph_copy()
    for node_id in g.nodes:
        app_log.debug(f"Processing node {node_id}")
        process_node(g.nodes[node_id])

    if tg.lattice_metadata:
        tg.lattice_metadata = encode_metadata(tg.lattice_metadata)

    tg_new._graph = g
    return tg_new


def process_lattice(lattice: Lattice) -> Lattice:
    """Convert a "legacy" (0.110.2) Lattice to a modern Lattice

    Args:
        lattice: old lattice

    Returns:
        the modernized lattice
    """

    workflow_function = lattice.workflow_function
    lattice.workflow_function = TransportableObject.make_transportable(workflow_function)
    inputs = {"args": lattice.args, "kwargs": lattice.kwargs}
    lattice.inputs = TransportableObject(inputs)

    workflow_function = lattice.workflow_function.get_deserialized()

    named_args, named_kwargs = get_named_params(workflow_function, lattice.args, lattice.kwargs)
    lattice.named_args = TransportableObject(named_args)
    lattice.named_kwargs = TransportableObject(named_kwargs)

    metadata = lattice.metadata

    if "workflow_executor" not in metadata:
        metadata["workflow_executor"] = "local"

    metadata = encode_metadata(metadata)
    lattice.metadata = metadata
    lattice.metadata["deps"] = {}
    lattice.metadata["call_before"] = []
    lattice.metadata["call_after"] = []

    lattice.transport_graph = process_transport_graph(lattice.transport_graph)
    lattice.transport_graph.lattice_metadata = lattice.metadata
    app_log.debug("Processed transport graph")

    # Delete raw inputs
    del lattice.__dict__["args"]
    del lattice.__dict__["kwargs"]

    return lattice


def process_result_object(result_object: Result) -> Result:
    """Convert a "legacy" (0.110.2) Result object to a modern Result object

    Args:
        result_object: the old Result object

    Returns:
        the modernized result object
    """

    app_log.debug(f"Processing result object for dispatch {result_object.dispatch_id}")
    process_lattice(result_object._lattice)
    app_log.debug("Processed lattice")

    result_object._result = TransportableObject.make_transportable(result_object._result)
    tg = result_object.lattice.transport_graph
    for n in tg._graph.nodes:
        tg.dirty_nodes.append(n)

    del result_object.__dict__["_inputs"]
    return result_object


def migrate_pickled_result_object(path: str) -> None:
    """Save legacy (0.110.2) result pickle file to a DataStore.

    This first transforms certain legacy properties of the result
    object and then persists the result object to the datastore.

    Args:
        path: path of the `result.pkl` file
    """

    with open(path, "rb") as f:
        result_object = pickle.load(f)

    process_result_object(result_object)
    update.persist(result_object)
